/*
 * cam.c
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/select.h>
#include <sys/time.h>
#include <linux/videodev.h>

#include "cam.h"

struct cam *cam_open(const char *dev_name)
{
    struct cam *c;
    struct video_capability vcap;
    struct video_picture vpict;

    /* Allocate camera struct.
     */
    c = malloc(sizeof(*c));
    if (!c) {
        fprintf(stderr, "%s: malloc(): %s\n", __func__, strerror(errno));
        goto err;
    }
    memset(c, 0, sizeof(*c));

    /* Open V4L device.
     */
    c->fd = open(dev_name, O_RDWR);
    if (c->fd < 0) {
        fprintf(stderr, "%s: open(\"%s\"): %s\n", __func__, dev_name, strerror(errno));
        goto err;
    }

    /* Get device capabilities.
     */
    if (ioctl(c->fd, VIDIOCGCAP, &vcap) < 0) {
        fprintf(stderr, "%s: ioctl(VIDIOCGCAP): %s\n", __func__, strerror(errno));
        goto err;
    }
    c->min_w = vcap.minwidth;
    c->min_h = vcap.minheight;
    c->max_w = vcap.maxwidth;
    c->max_h = vcap.maxheight;

    /* Set greyscale pixel format.
     */
    if (ioctl(c->fd, VIDIOCGPICT, &vpict) < 0) {
        fprintf(stderr, "%s: ioctl(VIDIOCGPICT): %s\n", __func__, strerror(errno));
        goto err;
    }
    vpict.depth = 8;
    vpict.palette = VIDEO_PALETTE_GREY;
    if (ioctl(c->fd, VIDIOCSPICT, &vpict) < 0) {
        fprintf(stderr, "%s: ioctl(VIDIOCSPICT): %s\n", __func__, strerror(errno));
        goto err;
    }

    /* Set maximum frame size.
     */
    if (cam_set_size(c, c->max_w, c->max_h))
        goto err;

    return c;

err:
    cam_close(c);
    return NULL;
}

void cam_close(struct cam *c)
{
    if (c) {
        cam_unmap_buffers(c);

        if (c->fd >= 0) {
            if (close(c->fd))
                fprintf(stderr, "%s: close(): %s\n", __func__, strerror(errno));
        }

        free(c);
    }
}

int cam_set_size(struct cam *c, unsigned w, unsigned h)
{
    struct video_window vwin;

    if (ioctl(c->fd, VIDIOCGWIN, &vwin) < 0) {
        fprintf(stderr, "%s: ioctl(VIDIOCGWIN): %s\n", __func__, strerror(errno));
        return -1;
    }

    vwin.width = w;
    vwin.height = h;

    if (ioctl(c->fd, VIDIOCSWIN, &vwin) < 0) {
        fprintf(stderr, "%s: ioctl(VIDIOCSWIN): %s\n", __func__, strerror(errno));
        return -1;
    }

    if (ioctl(c->fd, VIDIOCGWIN, &vwin) < 0) {
        fprintf(stderr, "%s: ioctl(VIDIOCGWIN): %s\n", __func__, strerror(errno));
        return -1;
    }

    c->w = vwin.width;
    c->h = vwin.height;

    return 0;
}

int cam_set_window(struct cam *c, unsigned x, unsigned y, unsigned w, unsigned h)
{
    struct video_capture vcapt;

    if (ioctl(c->fd, VIDIOCGCAPTURE, &vcapt) < 0) {
        fprintf(stderr, "%s: ioctl(VIDIOCGCAPTURE): %s\n", __func__, strerror(errno));
        return -1;
    }

    vcapt.x = x;
    vcapt.y = y;
    vcapt.width = w;
    vcapt.height = h;

    if (ioctl(c->fd, VIDIOCSCAPTURE, &vcapt) < 0) {
        fprintf(stderr, "%s: ioctl(VIDIOCSCAPTURE): %s\n", __func__, strerror(errno));
        return -1;
    }

    if (ioctl(c->fd, VIDIOCGCAPTURE, &vcapt) < 0) {
        fprintf(stderr, "%s: ioctl(VIDIOCGCAPTURE): %s\n", __func__, strerror(errno));
        return -1;
    }

    c->x = vcapt.x;
    c->y = vcapt.y;
    c->w = vcapt.width;
    c->h = vcapt.height;

    return 0;
}

int cam_read_frame(struct cam *c, unsigned char *buf)
{
    int result;
    int size = c->w * c->h;

    result = read(c->fd, buf, size);
    if (result < 0) {
        fprintf(stderr, "%s: read(): %s\n", __func__, strerror(errno));
        return -1;
    }
    if (result != size) {
        fprintf(stderr, "%s: read(): %d bytes expected, but %d bytes returned\n", __func__, size,
                result);
        return -1;
    }
    return 0;
}

int cam_map_buffers(struct cam *c)
{
    struct video_mbuf vmbuf;
    unsigned i;

    /* Get memory mapping info.
     */
    if (ioctl(c->fd, VIDIOCGMBUF, &vmbuf) < 0) {
        fprintf(stderr, "%s: ioctl(VIDIOCGMBUF): %s\n", __func__, strerror(errno));
        goto err;
    }

    /* Map device memory. Returned address is never NULL.
     */
    c->memsize = vmbuf.size;
    c->mem = mmap(NULL, c->memsize, PROT_READ|PROT_WRITE, MAP_SHARED, c->fd, 0);
    if (c->mem == MAP_FAILED) {
        fprintf(stderr, "%s: mmap(): %s\n", __func__, strerror(errno));
        goto err;
    }

    /* Allocate buffers array.
     */
    c->nbufs = vmbuf.frames;
    c->bufs = malloc(c->nbufs * sizeof(*c->bufs));
    if (!c->bufs) {
        fprintf(stderr, "%s: malloc(): %s\n", __func__, strerror(errno));
        goto err;
    }

    /* And initialize it.
     */
    for (i = 0; i < c->nbufs; i++)
        c->bufs[i] = c->mem + vmbuf.offsets[i];

    return 0;

err:
    cam_unmap_buffers(c);
    return -1;
}

void cam_unmap_buffers(struct cam *c)
{
    if (c->bufs)
        free(c->bufs);

    if (c->mem) {
        if (munmap(c->mem, c->memsize))
            fprintf(stderr, "%s: munmap(): %s\n", __func__, strerror(errno));
    }
}

int cam_capture_request(struct cam *c, unsigned buf_nr)
{
    struct video_mmap vmmap;

    vmmap.frame = buf_nr;
    vmmap.width = c->w;
    vmmap.height = c->h;
    vmmap.format = VIDEO_PALETTE_GREY;

    if (ioctl(c->fd, VIDIOCMCAPTURE, &vmmap) < 0) {
        fprintf(stderr, "%s: ioctl(VIDIOCMCAPTURE): %s\n", __func__, strerror(errno));
        return -1;
    }

    return 0;
}

int cam_capture_wait(struct cam *c, unsigned buf_nr)
{
    if (ioctl(c->fd, VIDIOCSYNC, &buf_nr) < 0) {
        fprintf(stderr, "%s: ioctl(VIDIOCSYNC): %s\n", __func__, strerror(errno));
        return -1;
    }
    return 0;
}

int cam_capture_ready(struct cam *c)
{
    fd_set set;
    struct timeval timeout;
    int result;

    FD_ZERO(&set);
    FD_SET(c->fd, &set);

    timeout.tv_sec = 0;
    timeout.tv_usec = 0;

    result = select(FD_SETSIZE, &set, NULL, NULL, &timeout);
    if (result < 0) {
        fprintf(stderr, "%s: select(): %s\n", __func__, strerror(errno));
        return -1;
    }

    return result;
}
