/*
 * sdlui.c
 *
 */

#include <stdlib.h>
#include <stdio.h>

#include <SDL/SDL.h>
#include <SDL/SDL_gfxPrimitives.h>

#include "sicam.h"
#include "sdlui.h"

double secPerFrame;

static SDL_Color grayColors[256];
static SDL_Surface *screen;
static struct cam *sicam;

static int exposure;
static int gain;
static int blackLevel;
static int blackBalance;

static int clockToInt(unsigned clock)
{
    switch (clock) {
        case SIUSB_CLK_20MHZ:   return 20;
        case SIUSB_CLK_25MHZ:   return 25;
        case SIUSB_CLK_30MHZ:   return 30;
        case SIUSB_CLK_35MHZ:   return 35;
        case SIUSB_CLK_40MHZ:   return 40;
        case SIUSB_CLK_45MHZ:   return 45;
        case SIUSB_CLK_50MHZ:   return 50;
        case SIUSB_CLK_55MHZ:   return 55;
        case SIUSB_CLK_60MHZ:   return 60;
        default:                fprintf(stderr, "%s: bad clock\n", __func__);
                                exit(1);
    }
}

static const char *shutterModeToStr(unsigned shutterMode)
{
    switch (shutterMode) {
        case SIUSB_SH_TRIG_TB1:     return "TRIG_TB1";
        case SIUSB_SH_TRIG_TB2:     return "TRIG_TB2";
        case SIUSB_SH_TRIG_TB3:     return "TRIG_TB3";
        case SIUSB_SH_TRIG_TB4:     return "TRIG_TB4";
        case SIUSB_SH_ROLLING:      return "ROLLING";
        case SIUSB_SH_ROLLING_DS1:  return "ROLLING_DS1";
        case SIUSB_SH_ROLLING_DS2:  return "ROLLING_DS2";
        case SIUSB_SH_ROLLING_DS3:  return "ROLLING_DS3";
        default:                    fprintf(stderr, "%s: bad shutter mode\n", __func__);
                                    exit(1);
    }
}

static int shutterModeToClocks(unsigned shutterMode)
{
    switch (shutterMode) {
        case SIUSB_SH_TRIG_TB1: return SIUSB_EXPCLK_TB1;
        case SIUSB_SH_TRIG_TB2: return SIUSB_EXPCLK_TB2;
        case SIUSB_SH_TRIG_TB3: return SIUSB_EXPCLK_TB3;
        case SIUSB_SH_TRIG_TB4: return SIUSB_EXPCLK_TB4;
        default:                fprintf(stderr, "%s: bad shutter mode\n", __func__);
                                exit(1);
    }
}

static double gainToDouble(unsigned gain)
{
    switch (gain) {
        case SIUSB_GAIN_1_41:   return 1.41;
        case SIUSB_GAIN_1_62:   return 1.62;
        case SIUSB_GAIN_1_88:   return 1.88;
        case SIUSB_GAIN_2_17:   return 2.17;
        case SIUSB_GAIN_2_52:   return 2.52;
        case SIUSB_GAIN_2_92:   return 2.92;
        case SIUSB_GAIN_3_38:   return 3.38;
        case SIUSB_GAIN_3_90:   return 3.90;
        case SIUSB_GAIN_4_50:   return 4.50;
        case SIUSB_GAIN_5_17:   return 5.17;
        case SIUSB_GAIN_5_93:   return 5.93;
        case SIUSB_GAIN_6_76:   return 6.76;
        case SIUSB_GAIN_7_86:   return 7.86;
        case SIUSB_GAIN_8_64:   return 8.64;
        case SIUSB_GAIN_9_61:   return 9.61;
        case SIUSB_GAIN_10_69:  return 10.69;
        case SIUSB_GAIN_1_00:   return 1.00;
        default:                fprintf(stderr, "%s: bad gain\n", __func__);
                                exit(1);
    }
}

void sdlUIInit(int w, int h, struct cam *c)
{
    unsigned i;

    sicam = c;

    exposure = sicam_get_exposure(sicam);
    gain = sicam_get_gain(sicam);
    blackLevel = sicam_get_black_level(sicam);
    blackBalance = sicam_get_black_balance(sicam);

    if (SDL_Init(SDL_INIT_VIDEO)) {
        fprintf(stderr, "%s: SDL_Init() failed: %s\n", __func__, SDL_GetError());
        exit(1);
    }

    if (atexit(SDL_Quit)) {
        fprintf(stderr, "%s: atexit() failed\n", __func__);
        SDL_Quit();
        exit(1);
    }

    for (i = 0; i < 256; i++) {
        grayColors[i].r = i;
        grayColors[i].g = i;
        grayColors[i].b = i;
    }

    screen = SDL_SetVideoMode(w, h, 8, SDL_HWSURFACE|SDL_HWPALETTE|SDL_DOUBLEBUF);
    if (!screen) {
        fprintf(stderr, "%s: SDL_SetVideoMode() failed: %s\n", __func__, SDL_GetError());
        exit(1);
    }

    SDL_SetPalette(screen, SDL_LOGPAL|SDL_PHYSPAL, grayColors, 0, 256);

    if (SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL)) {
        fprintf(stderr, "%s: SDL_EnableKeyRepeat() failed: %s\n", __func__, SDL_GetError());
        exit(1);
    }
}

void sdlUIBlit(unsigned char *buf)
{
    char tmp[100];
    SDL_Surface *image;
    int clock;
    unsigned shutterMode;
    double exposureUSec;

    image = SDL_CreateRGBSurfaceFrom(buf, screen->w, screen->h, 8, screen->w, 0, 0, 0, 0);
    if (!image) {
        fprintf(stderr, "%s: SDL_CreateRGBSurfaceFrom() failed: %s\n", __func__, SDL_GetError());
        exit(1);
    }

    SDL_SetPalette(image, SDL_LOGPAL|SDL_PHYSPAL, grayColors, 0, 256);

    if (SDL_BlitSurface(image, NULL, screen, NULL)) {
        fprintf(stderr, "%s: SDL_BlitSurface() failed: %s\n", __func__, SDL_GetError());
        exit(1);
    }

    SDL_FreeSurface(image);

    clock =  clockToInt(sicam_get_clock(sicam));
    shutterMode = sicam_get_shutter_mode(sicam);

    sprintf(tmp, "Clock:          %2d MHz", clock);
    stringRGBA(screen, screen->w-250, 10, tmp, 255, 255, 255, 255);

    sprintf(tmp, "Shutter Mode:   %11s", shutterModeToStr(shutterMode));
    stringRGBA(screen, screen->w-250, 20, tmp, 255, 255, 255, 255);

    sprintf(tmp, "Exposure Count: %4d", exposure);
    stringRGBA(screen, screen->w-250, 30, tmp, 255, 255, 255, 255);

    sprintf(tmp, "Gain:           %5.2f", gainToDouble(sicam_get_gain(sicam)));
    stringRGBA(screen, screen->w-250, 40, tmp, 255, 255, 255, 255);

    sprintf(tmp, "Black Offset:   %3d", blackLevel);
    stringRGBA(screen, screen->w-250, 50, tmp, 255, 255, 255, 255);

    sprintf(tmp, "Column Offset:  %3d", blackBalance);
    stringRGBA(screen, screen->w-250, 60, tmp, 255, 255, 255, 255);

    switch (shutterMode) {
        case SIUSB_SH_TRIG_TB1:
        case SIUSB_SH_TRIG_TB2:
        case SIUSB_SH_TRIG_TB3:
        case SIUSB_SH_TRIG_TB4:     exposureUSec = (double)(exposure * shutterModeToClocks(shutterMode)) / clock;
                                    break;
        // TODO: Seems that I don't calculate this correct!
        case SIUSB_SH_ROLLING:
        case SIUSB_SH_ROLLING_DS1:
        case SIUSB_SH_ROLLING_DS2:
        case SIUSB_SH_ROLLING_DS3:  exposureUSec = (double)((sicam->w + 150) * clock) * (sicam->h - exposure) / 1000000;
                                    break;
        default:                    fprintf(stderr, "%s: bad shutter mode\n", __func__);
                                    exit(1);
    }

    sprintf(tmp, "Exposure Time:  %8.2f us", exposureUSec);
    stringRGBA(screen, screen->w-250, 80, tmp, 255, 255, 255, 255);

    sprintf(tmp, "FPS:            %4.1f", 1.0 / secPerFrame);
    stringRGBA(screen, screen->w-250, 100, tmp, 255, 255, 255, 255);

    if (SDL_Flip(screen)) {
        fprintf(stderr, "%s: SDL_Flip() failed: %s\n", __func__, SDL_GetError());
        exit(1);
    }
}

int sdlUIEvents(void)
{
    SDL_Event event;
    int quit = 0;

    while (SDL_PollEvent(&event)) {
        switch (event.type) {
            case SDL_KEYDOWN:
                switch (event.key.keysym.sym) {
                    case SDLK_ESCAPE:
                        quit = 1;
                        break;

                    case SDLK_f:
                        if (!SDL_WM_ToggleFullScreen(screen))
                            fprintf(stderr, "%s: SDL_WM_ToggleFullScreen() failed: %s\n", __func__,
                                    SDL_GetError());
                        break;

                    /* Clock
                     */

                    case SDLK_q:
                        sicam_set_clock(sicam, SIUSB_CLK_20MHZ);
                        break;

                    case SDLK_w:
                        sicam_set_clock(sicam, SIUSB_CLK_25MHZ);
                        break;

                    case SDLK_e:
                        sicam_set_clock(sicam, SIUSB_CLK_30MHZ);
                        break;

                    case SDLK_r:
                        sicam_set_clock(sicam, SIUSB_CLK_35MHZ);
                        break;

                    case SDLK_t:
                        sicam_set_clock(sicam, SIUSB_CLK_40MHZ);
                        break;

                    case SDLK_y:
                        sicam_set_clock(sicam, SIUSB_CLK_45MHZ);
                        break;

                    case SDLK_u:
                        sicam_set_clock(sicam, SIUSB_CLK_50MHZ);
                        break;

                    case SDLK_i:
                        sicam_set_clock(sicam, SIUSB_CLK_55MHZ);
                        break;

                    case SDLK_o:
                        sicam_set_clock(sicam, SIUSB_CLK_60MHZ);
                        break;

                    /* Shutter mode
                     */

                    case SDLK_1:
                        sicam_set_shutter_mode(sicam, SIUSB_SH_TRIG_TB1);
                        break;

                    case SDLK_2:
                        sicam_set_shutter_mode(sicam, SIUSB_SH_TRIG_TB2);
                        break;

                    case SDLK_3:
                        sicam_set_shutter_mode(sicam, SIUSB_SH_TRIG_TB3);
                        break;

                    case SDLK_4:
                        sicam_set_shutter_mode(sicam, SIUSB_SH_TRIG_TB4);
                        break;

                    case SDLK_5:
                        sicam_set_shutter_mode(sicam, SIUSB_SH_ROLLING);
                        break;

                    case SDLK_6:
                        sicam_set_shutter_mode(sicam, SIUSB_SH_ROLLING_DS1);
                        break;

                    case SDLK_7:
                        sicam_set_shutter_mode(sicam, SIUSB_SH_ROLLING_DS2);
                        break;

                    case SDLK_8:
                        sicam_set_shutter_mode(sicam, SIUSB_SH_ROLLING_DS3);
                        break;

                    /* Exposure
                     */

                    case SDLK_F5:
                        if (event.key.keysym.mod & KMOD_SHIFT)
                            exposure -= 1;
                        else
                            exposure -= 100;
                        if (exposure < 0)
                            exposure = 0;
                        sicam_set_exposure(sicam, exposure);
                        break;

                    case SDLK_F6:
                        if (event.key.keysym.mod & KMOD_SHIFT)
                            exposure += 1;
                        else
                            exposure += 100;
                        if (exposure > SIUSB_EXPOSURE_MAX)
                            exposure = SIUSB_EXPOSURE_MAX;
                        sicam_set_exposure(sicam, exposure);
                        break;

                    /* Gain
                     */

                    case SDLK_F7:
                        gain--;
                        if (gain < SIUSB_GAIN_MIN)
                            gain = SIUSB_GAIN_MIN;
                        sicam_set_gain(sicam, gain);
                        break;

                    case SDLK_F8:
                        gain++;
                        if (gain > SIUSB_GAIN_UNITY)
                            gain = SIUSB_GAIN_UNITY;
                        sicam_set_gain(sicam, gain);
                        break;

                    /* Black level
                     */

                    case SDLK_F9:
                        blackLevel -= 5;
                        if (blackLevel < 0)
                            blackLevel = 0;
                        sicam_set_black_level(sicam, blackLevel);
                        break;

                    case SDLK_F10:
                        blackLevel += 5;
                        if (blackLevel > SIUSB_BLACK_LEVEL_MAX)
                            blackLevel = SIUSB_BLACK_LEVEL_MAX;
                        sicam_set_black_level(sicam, blackLevel);
                        break;

                    /* Black balance
                     */

                    case SDLK_F11:
                        blackBalance -= 5;
                        if (blackBalance < 0)
                            blackBalance = 0;
                        sicam_set_black_balance(sicam, blackBalance);
                        break;

                    case SDLK_F12:
                        blackBalance += 5;
                        if (blackBalance > SIUSB_BLACK_BALANCE_MAX)
                            blackBalance = SIUSB_BLACK_BALANCE_MAX;
                        sicam_set_black_balance(sicam, blackBalance);
                        break;

                    default:
                        break;
                }
                break;

            case SDL_QUIT:
                quit = 1;
                break;
        }
    }
    return quit;
}
