/*
 * siusb.c
 *
 * SI-1280F USB Camera Driver
 *
 * Copyright (C) 2003-2004 Nikka Research Deutschland GmbH
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/poll.h>
#include <linux/usb.h>
#include <asm/uaccess.h>
#include <linux/videodev.h>

#include "siusb.h"


/* Kernel compatibility
 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#define LINUX_24
#endif

#ifdef LINUX_24
/* 2.4 does not have the mem_flags argument */
#define siusb_submit_urb(urb, mem_flags)    usb_submit_urb(urb)
#define siusb_alloc_urb(urb, mem_flags)     usb_alloc_urb(urb)
#else /* !LINUX_24 */
#define siusb_submit_urb(urb, mem_flags)    usb_submit_urb(urb, mem_flags)
#define siusb_alloc_urb(urb, mem_flags)     usb_alloc_urb(urb, mem_flags)
#endif /* !LINUX_24 */


/* Debug macros
 */
#ifdef DEBUG
#define TRACE1(fmt, args...)    do { if (DEBUG >= 1) printk(KERN_DEBUG fmt, ## args); } while (0)
#define TRACE2(fmt, args...)    do { if (DEBUG >= 2) printk(KERN_DEBUG fmt, ## args); } while (0)
#define TRACE3(fmt, args...)    do { if (DEBUG >= 3) printk(KERN_DEBUG fmt, ## args); } while (0)
#else /* !DEBUG */
#define TRACE1(fmt, args...)    do { } while (0)
#define TRACE2(fmt, args...)    do { } while (0)
#define TRACE3(fmt, args...)    do { } while (0)
#endif /* !DEBUG */


#define DRIVER_AUTHOR   "Nikka Research Deutschland GmbH"
#define DRIVER_DESC     "SI-1280F USB Camera Driver"
#define DRIVER_VERSION  "v0.8.4"


/* Module parameters
 */

#ifndef SIUSB_NUM_BUFS
#define SIUSB_NUM_BUFS      2
#endif

static unsigned num_bufs = SIUSB_NUM_BUFS;

MODULE_PARM(num_bufs, "i");
MODULE_PARM_DESC(num_bufs,
    "Number of buffers (default is " __stringify(SIUSB_NUM_BUFS) ")");

#ifndef SIUSB_NUM_BULK_URBS
#define SIUSB_NUM_BULK_URBS 32
#endif

static unsigned num_bulk_urbs = SIUSB_NUM_BULK_URBS;

MODULE_PARM(num_bulk_urbs, "i");
MODULE_PARM_DESC(num_bulk_urbs,
    "Number of bulk URBs (default is " __stringify(SIUSB_NUM_BULK_URBS) ")");


#define SIUSB_PACKET_SIZE   512

#define SIUSB_VERSION_SIZE  16
#define SIUSB_CPLD_SIZE     4


#define SIUSB_MIN_WIDTH     256
#define SIUSB_MIN_HEIGHT    1

#define SIUSB_MAX_WIDTH     1280
#define SIUSB_MAX_HEIGHT    1024

#define SIUSB_MAX_SIZE      (SIUSB_MAX_WIDTH * SIUSB_MAX_HEIGHT)


/* Numer of pages in pages array of each buffer. This must be a whole number of
 * pages, i.e. SIUSB_MAX_SIZE % PAGE_SIZE must be 0!
 */
#define SIUSB_PAGES_PER_BUF (SIUSB_MAX_SIZE / PAGE_SIZE)


#define SIUSB_URB_TIMEOUT   (HZ/2)  /* Half a second */


/* Arm trig, origin, width, height, clock, shutter mode, exposure, gain, black
 * level, black balance, disarm trig -- total 11.
 */
#define SIUSB_MAX_VRS       11


#ifndef VID_HARDWARE_SI
#define VID_HARDWARE_SI     71232   /* Random */
#endif


/* Buffer states
 */
#define SIUSB_BUF_UNUSED    0   /* Unused (no MCAPTURE) */
#define SIUSB_BUF_QUEUED    1   /* In capture queue */
#define SIUSB_BUF_DONE      2   /* Finished grabbing, but not been synced yet */


struct siusb_buf {
    unsigned nr;                /* Index in buffers array */
    unsigned state;             /* Buffer state */
    int result;                 /* Negative error code or 0 = no error */

    unsigned trig_mode;         /* Internal or external trigger mode */

    /* Requested register settings for this frame
     */
    unsigned x, y, w, h;
    unsigned clock;
    unsigned shutter_mode;
    unsigned exposure;
    unsigned gain;
    unsigned black_level;
    unsigned black_balance;

    unsigned long *pages;       /* Pages array */
    wait_queue_head_t done;     /* Wait for readout finish */
    struct list_head captq;     /* Capture queue list */
};


struct siusb_vendor_req {
    const char *name;           /* Name, for error reporting and debug purposes */

    /* Vendor request info
     */
    __u8 dir;
    __u8 request;
    __u16 value;
    __u16 index;
    void *data;
    __u16 size;

    int timeout;                /* urb->timeout */

    unsigned *phwval1;          /* On transfer completion we set *phwvalX = newvalX */
    unsigned newval1;
    unsigned *phwval2;
    unsigned newval2;
};


struct siusb {
    struct video_device *vdev;  /* V4L device for this camera, NULL means cam disconnected */
    struct usb_device *udev;    /* Save USB device pointer */
    int open_count;             /* Device open count */
    struct semaphore sem;       /* Locks this structure */
    int bulk_ep_addr;           /* Bulk IN endpoint address */

    char version[SIUSB_VERSION_SIZE];   /* Keep camera version string */

    struct siusb_buf *bufs;     /* Buffers array */
    struct siusb_buf read_buf;  /* read() syscall buffer */

    int switched_to_mmap;       /* True if switched to using mmap() instead of read() */

    struct list_head captq;     /* Capture queue */
    spinlock_t captq_lock;      /* Lock capture queue add/remove */

    struct usb_ctrlrequest vr_setup;            /* Setup packet for vendor requests */
    int nvrs;                                   /* Number of requests in vrs array */
    struct siusb_vendor_req vrs[SIUSB_MAX_VRS]; /* Vendor requests to perform before capture */

    struct urb *init_urb;       /* Used to initiate readout */
    struct urb **bulk_urbs;     /* Queue of bulk transfer requests */

    int nr_req;                 /* Number of queued bulk transfer requests */
    int rem;                    /* Remaining bytes */
    unsigned cpage;             /* Current page in pages array */

    unsigned trig_mode;         /* Trigger mode */

    /* Current in-hardware register settings
     */
    unsigned hw_x, hw_y, hw_w, hw_h;
    unsigned hw_clock;
    unsigned hw_shutter_mode;
    unsigned hw_exposure;
    unsigned hw_gain;
    unsigned hw_black_level;
    unsigned hw_black_balance;

    /* Requested register settings
     */
    unsigned x, y, w, h;
    unsigned clock;
    unsigned shutter_mode;
    unsigned exposure;
    unsigned gain;
    unsigned black_level;
    unsigned black_balance;
};


static DECLARE_MUTEX(disconnect_sem);   /* Prevent races between open() and disconnect() */


/*
 * Low-level
 */

/* Vendor request codes
 */
#define SIUSB_REQ_VERSION           0xE1    /* Read firmware version */
#define SIUSB_REQ_CPLD              0xE8    /* Read CPLD version */
#define SIUSB_REQ_FLASH             0xE4    /* Read/Write flash memory */
#define SIUSB_REQ_CLOCK             0xE2    /* Get/Set clock */
#define SIUSB_REQ_LOAD_REG          0xEF    /* Load sensor registers */
#define SIUSB_REQ_SET_ORIGIN        0xE6    /* Set row/column origin (pan/tilt) */
#define SIUSB_REQ_SET_DEPTH         0xED    /* Set bit depth (8 or 16 bits) */
#define SIUSB_REQ_READOUT           0xEA    /* Snapshot expose & readout */
#define SIUSB_REQ_ARM_TRIG          0xEC    /* Arm triggered, snapshot exposure mode */
#define SIUSB_REQ_DISARM_TRIG       0xEB    /* Disarm triggered mode */

/* Sensor registers
 */
#define SIUSB_REG_SHUTTER_MODE      0x00
#define SIUSB_REG_WIDTH             0x01    /* (Pixels in row) / 2 - 1 */
#define SIUSB_REG_HEIGHT            0x02    /* Number of rows - 1 */
#define SIUSB_REG_EXPOSURE          0x03
#define SIUSB_REG_BEGIN_COLUMN      0x04    /* X-origin / 2 */
#define SIUSB_REG_BEGIN_ROW         0x05    /* Y-origin */
#define SIUSB_REG_BEGIN_ROW_2       0x06    /* Set to same value as Reg5 */
#define SIUSB_REG_SUBSAMPLING       0x07
#define SIUSB_REG_GAIN              0x08
#define SIUSB_REG_BLACK_LEVEL       0x09
#define SIUSB_REG_BLACK_BALANCE     0x0A

/* Subsampling
 */
#define SIUSB_SUBSAMP_NONE          0x00
#define SIUSB_SUBSAMP_2X            0x6C
#define SIUSB_SUBSAMP_2X_BAYER      0x24

/* Bit depth
 */
#define SIUSB_DEPTH_8BIT            0
#define SIUSB_DEPTH_16BIT           1


static int siusb_ctrl_msg(struct siusb *s, __u8 dir, __u8 request,
                          __u16 value, __u16 index, void *data, __u16 size)
{
    return usb_control_msg(s->udev,
                           dir ? usb_rcvctrlpipe(s->udev, 0) : usb_sndctrlpipe(s->udev, 0),
                           request, dir | USB_TYPE_VENDOR | USB_RECIP_ENDPOINT,
                           value, index, data, size, SIUSB_URB_TIMEOUT);
}

static int siusb_read_version(struct siusb *s, char *buf)
{
    int result;

    result = siusb_ctrl_msg(s, USB_DIR_IN, SIUSB_REQ_VERSION, 0, 0, buf, SIUSB_VERSION_SIZE);
    if (result != SIUSB_VERSION_SIZE) {
        printk(KERN_ERR "siusb: %s()=%d: FAILED\n", __func__, result);
        if (result >= 0)
            result = -EIO;
    } else {
        /* Be sure we return with properly null-terminated buf.
         */
        buf[SIUSB_VERSION_SIZE-1] = '\0';
    }
    return result;
}

static int siusb_read_cpld(struct siusb *s, char *buf)
{
    int result;

    result = siusb_ctrl_msg(s, USB_DIR_IN, SIUSB_REQ_CPLD, 0, 0, buf, SIUSB_CPLD_SIZE);
    if (result != SIUSB_CPLD_SIZE) {
        printk(KERN_ERR "siusb: %s()=%d: FAILED\n", __func__, result);
        if (result >= 0)
            result = -EIO;
    }
    return result;
}

static int siusb_set_subsampling(struct siusb *s, unsigned subsamp)
{
    int result;

    result = siusb_ctrl_msg(s, USB_DIR_IN, SIUSB_REQ_LOAD_REG, subsamp, SIUSB_REG_SUBSAMPLING,
                            NULL, 0);
    if (result < 0)
        printk(KERN_ERR "siusb: %s(subsamp=0x%02X)=%d: FAILED\n", __func__, subsamp, result);
    return result;
}

static int siusb_set_depth(struct siusb *s, unsigned depth)
{
    int result;

    result = siusb_ctrl_msg(s, USB_DIR_IN, SIUSB_REQ_SET_DEPTH, depth, 0, NULL, 0);
    if (result < 0)
        printk(KERN_ERR "siusb: %s(depth=%u)=%d: FAILED\n", __func__, depth, result);
    return result;
}


/*
 * Capture queue
 */

static void siusb_report_urb_status(struct siusb *s, struct siusb_buf *buf, struct urb *urb)
{
    TRACE3("siusb: buf#%u: %s %p: status=%d, req %d, rcv %d\n", buf->nr,
           urb == s->init_urb ? "INIT" : "BULK", urb, urb->status,
           urb->transfer_buffer_length, urb->actual_length);

    /* Report failed URBs, but be silent on unlinks (-ENOENT).
     */
    if (urb->status < 0) {
#ifndef DEBUG
        if (urb->status == -ENOENT)
            return;
#endif
        printk(KERN_ERR "siusb: buf#%u: %s %p: status=%d, req %d, rcv %d: FAILED\n", buf->nr,
               urb == s->init_urb ? "INIT" : "BULK", urb, urb->status,
               urb->transfer_buffer_length, urb->actual_length);
    }
}

static void siusb_capture_start(struct siusb *s);

static void siusb_buf_complete(struct siusb *s, struct siusb_buf *buf)
{
    unsigned long flags;
    unsigned start_capture = 0;

    /* Frame transfer is complete, when all queued URBs are complete.
     */
    if (s->nr_req > 0)
        return;

    TRACE2("siusb: buf#%u: DONE, result=%d\n", buf->nr, buf->result);

    spin_lock_irqsave(&s->captq_lock, flags);

    list_del(&buf->captq);  /* Remove from capture queue */

    if (!list_empty(&s->captq)) {
        TRACE2("siusb: %s: buf#%u: going to start capture\n", __func__, buf->nr);
        start_capture = 1;
    }

    spin_unlock_irqrestore(&s->captq_lock, flags);

    buf->state = SIUSB_BUF_DONE;
    TRACE2("siusb: buf#%u: wakeing up\n", buf->nr);
    wake_up(&buf->done);

    if (start_capture)
        siusb_capture_start(s);
}

/* Returns 1 if transfer should be completed (either successfully or because of
 * an error).
 */
static int siusb_submit_bulk_urb(struct siusb *s, struct urb *urb, struct siusb_buf *buf)
{
    if (s->rem <= 0)
        return 1;

    urb->transfer_buffer = (void *)buf->pages[s->cpage];
    urb->transfer_buffer_length = min(s->rem, (int)PAGE_SIZE);
    buf->result = siusb_submit_urb(urb, GFP_ATOMIC);
    if (buf->result < 0) {
        printk(KERN_ERR "siusb: buf#%u: usb_submit_urb(BULK %p)=%d: FAILED\n", buf->nr, urb,
               buf->result);
        return 1;
    }
    TRACE2("siusb: buf#%u: BULK %p submitted, length=%d, rem=%d\n", buf->nr,
           urb, urb->transfer_buffer_length, s->rem);

    s->cpage++;
    s->rem -= PAGE_SIZE;

    s->nr_req++;
    return 0;
}

static void siusb_init_urb_cb(struct urb *urb
#ifndef LINUX_24
, struct pt_regs *regs
#endif
)
{
    struct siusb *s = urb->context;
    struct siusb_buf *buf = list_entry(s->captq.next, struct siusb_buf, captq);
    unsigned i;

    siusb_report_urb_status(s, buf, urb);

    buf->result = urb->status;
    if (buf->result < 0)
        goto complete;

    for (i = 0; i < num_bulk_urbs; i++) {
        if (siusb_submit_bulk_urb(s, s->bulk_urbs[i], buf))
            goto complete;
    }

    return;

complete:
    siusb_buf_complete(s, buf);
}

static void siusb_bulk_urb_cb(struct urb *urb
#ifndef LINUX_24
, struct pt_regs *regs
#endif
)
{
    struct siusb *s = urb->context;
    struct siusb_buf *buf = list_entry(s->captq.next, struct siusb_buf, captq);

    siusb_report_urb_status(s, buf, urb);

    s->nr_req--;

    if (buf->result < 0)
        goto complete;

    buf->result = urb->status;
    if (buf->result < 0)
        goto complete;

    if (siusb_submit_bulk_urb(s, urb, buf))
        goto complete;

    return;

complete:
    siusb_buf_complete(s, buf);
}

static int siusb_readout_start(struct siusb *s, struct siusb_buf *buf)
{
    int result;

    /* Submit readout start URB. Transfer length should be 1 byte, we're using
     * 8-bit depth.
     */
    s->init_urb->timeout = SIUSB_URB_TIMEOUT;
    usb_fill_bulk_urb(s->init_urb, s->udev, usb_rcvbulkpipe(s->udev, s->bulk_ep_addr),
                      (void *)buf->pages[0], 1, siusb_init_urb_cb, s);
    result = siusb_submit_urb(s->init_urb, GFP_ATOMIC);
    if (result < 0) {
        printk(KERN_ERR "siusb: buf#%u: usb_submit_urb(INIT)=%d: FAILED\n", buf->nr, result);
        return result;
    }
    TRACE2("siusb: buf#%u: INIT submitted\n", buf->nr);
    return 0;
}

static int siusb_submit_next_vr(struct siusb *s, struct siusb_buf *buf);

static void siusb_vendor_req_cb(struct urb *urb
#ifndef LINUX_24
, struct pt_regs *regs
#endif
)
{
    struct siusb *s = urb->context;
    struct siusb_buf *buf = list_entry(s->captq.next, struct siusb_buf, captq);
    struct siusb_vendor_req *vr = &s->vrs[s->nvrs];

    siusb_report_urb_status(s, buf, urb);

    buf->result = urb->status;
    if (buf->result < 0) {
#ifndef DEBUG
        if (buf->result == -ENOENT) /* Be silent for unlinked URBs */
            goto complete;
#endif
        printk(KERN_ERR "siusb: buf#%u: %s: dir=%s; value=0x%04X; index=0x%04X\n",
               buf->nr, vr->name, vr->dir ? "IN" : "OUT", vr->value, vr->index);
        goto complete;
    }

    /* Vendor request completed successfully, so new setting is in charge.
     */
    if (vr->phwval1)
        *vr->phwval1 = vr->newval1;
    if (vr->phwval2)
        *vr->phwval2 = vr->newval2;

    buf->result = siusb_submit_next_vr(s, buf);
    if (buf->result < 0)
        goto complete;

    return;

complete:
    siusb_buf_complete(s, buf);
}

static int siusb_submit_vendor_req(struct siusb *s, const struct siusb_vendor_req *vr)
{
    /* Fill setup packet for vendor request.
     */
    s->vr_setup.bRequestType = vr->dir | USB_TYPE_VENDOR | USB_RECIP_ENDPOINT,
    s->vr_setup.bRequest = vr->request;
    s->vr_setup.wValue = cpu_to_le16(vr->value);
    s->vr_setup.wIndex = cpu_to_le16(vr->index);
    s->vr_setup.wLength = cpu_to_le16(vr->size);

    /* TODO:
     * Timeout value in URBs means nothing. Seems like the USB core does not
     * support this yet. Wait for USB core to support timeout for URBs or
     * implement it ourselfs?
     */

    s->init_urb->timeout = vr->timeout;
    usb_fill_control_urb(s->init_urb, s->udev,
                         vr->dir ? usb_rcvctrlpipe(s->udev, 0) : usb_sndctrlpipe(s->udev, 0),
                         (unsigned char *)&s->vr_setup, vr->data, vr->size,
                         siusb_vendor_req_cb, s);

    return siusb_submit_urb(s->init_urb, GFP_ATOMIC);
}

static int siusb_submit_next_vr(struct siusb *s, struct siusb_buf *buf)
{
    int result = 0;

    /* If there are more vendor requests remaining, submit the next one, else
     * start image readout.
     */
    if (s->nvrs > 0) {
        result = siusb_submit_vendor_req(s, &s->vrs[--s->nvrs]);
        if (result < 0)
            printk(KERN_ERR "siusb: buf#%u: siusb_submit_vendor_req(%s)=%d: FAILED\n", buf->nr,
                   s->vrs[s->nvrs].name, result);
        TRACE2("siusb: buf#%u: %s submitted\n", buf->nr, s->vrs[s->nvrs].name);
    } else
        result = siusb_readout_start(s, buf);
    return result;
}

static void siusb_fill_vendor_req(struct siusb_vendor_req *vr, const char *name, __u8 dir,
                                  __u8 request, __u16 value, __u16 index, void *data, __u16 size,
                                  int timeout,
                                  unsigned *phwval1, unsigned newval1,
                                  unsigned *phwval2, unsigned newval2)
{
    vr->name = name;

    vr->dir = dir;
    vr->request = request;
    vr->value = value;
    vr->index = index;
    vr->data = data;
    vr->size = size;

    vr->timeout = timeout;

    vr->phwval1 = phwval1;
    vr->newval1 = newval1;
    vr->phwval2 = phwval2;
    vr->newval2 = newval2;
}

static void siusb_fill_vrs(struct siusb *s, struct siusb_buf *buf)
{
    /* Fill a table of vendor requests to be submitted before actual frame
     * transfer. Vendor requests are submitted in reverse order, i.e. the first
     * request to be submitted should be added to the table last here.
     */

    s->nvrs = 0;

    /* The last request to be submitted is the one to wait for the external
     * trigger, if capture by external trigger is selected. We may wait for
     * external trigger indefinitely, so set timeout to 0.
     */
    if (buf->trig_mode)
        siusb_fill_vendor_req(&s->vrs[s->nvrs++], "ARM TRIG", USB_DIR_IN, SIUSB_REQ_ARM_TRIG,
                              0, 0, NULL, 0,
                              0,
                              NULL, 0, NULL, 0);

    if (s->hw_black_balance != buf->black_balance)
        siusb_fill_vendor_req(&s->vrs[s->nvrs++], "BLACK BALANCE", USB_DIR_IN, SIUSB_REQ_LOAD_REG,
                              buf->black_balance, SIUSB_REG_BLACK_BALANCE, NULL, 0,
                              SIUSB_URB_TIMEOUT,
                              &s->hw_black_balance, buf->black_balance, NULL, 0);

    if (s->hw_black_level != buf->black_level)
        siusb_fill_vendor_req(&s->vrs[s->nvrs++], "BLACK LEVEL", USB_DIR_IN, SIUSB_REQ_LOAD_REG,
                              buf->black_level, SIUSB_REG_BLACK_LEVEL, NULL, 0,
                              SIUSB_URB_TIMEOUT,
                              &s->hw_black_level, buf->black_level, NULL, 0);

    if (s->hw_gain != buf->gain)
        siusb_fill_vendor_req(&s->vrs[s->nvrs++], "GAIN", USB_DIR_IN, SIUSB_REQ_LOAD_REG,
                              buf->gain, SIUSB_REG_GAIN, NULL, 0,
                              SIUSB_URB_TIMEOUT,
                              &s->hw_gain, buf->gain, NULL, 0);

    if (s->hw_exposure != buf->exposure)
        siusb_fill_vendor_req(&s->vrs[s->nvrs++], "EXPOSURE", USB_DIR_IN, SIUSB_REQ_LOAD_REG,
                              buf->exposure, SIUSB_REG_EXPOSURE, NULL, 0,
                              SIUSB_URB_TIMEOUT,
                              &s->hw_exposure, buf->exposure, NULL, 0);

    if (s->hw_shutter_mode != buf->shutter_mode)
        siusb_fill_vendor_req(&s->vrs[s->nvrs++], "SHUTTER MODE", USB_DIR_IN, SIUSB_REQ_LOAD_REG,
                              buf->shutter_mode, SIUSB_REG_SHUTTER_MODE, NULL, 0,
                              SIUSB_URB_TIMEOUT,
                              &s->hw_shutter_mode, buf->shutter_mode, NULL, 0);

    if (s->hw_clock != buf->clock)
        siusb_fill_vendor_req(&s->vrs[s->nvrs++], "CLOCK", USB_DIR_OUT, SIUSB_REQ_CLOCK,
                              0, 0, &buf->clock, 3,
                              SIUSB_URB_TIMEOUT,
                              &s->hw_clock, buf->clock, NULL, 0);

    if (s->hw_h != buf->h)
        siusb_fill_vendor_req(&s->vrs[s->nvrs++], "HEIGHT", USB_DIR_IN, SIUSB_REQ_LOAD_REG,
                              buf->h-1, SIUSB_REG_HEIGHT, NULL, 0,
                              SIUSB_URB_TIMEOUT,
                              &s->hw_h, buf->h, NULL, 0);

    if (s->hw_w != buf->w)
        siusb_fill_vendor_req(&s->vrs[s->nvrs++], "WIDTH", USB_DIR_IN, SIUSB_REQ_LOAD_REG,
                              buf->w/2-1, SIUSB_REG_WIDTH, NULL, 0,
                              SIUSB_URB_TIMEOUT,
                              &s->hw_w, buf->w, NULL, 0);

    if (s->hw_x != buf->x || s->hw_y != buf->y)
        siusb_fill_vendor_req(&s->vrs[s->nvrs++], "ORIGIN", USB_DIR_IN, SIUSB_REQ_SET_ORIGIN,
                              buf->x, buf->y, NULL, 0,
                              SIUSB_URB_TIMEOUT,
                              &s->hw_x, buf->x, &s->hw_y, buf->y);

    /* TODO: Sometimes after switching from external to internal trigger mode,
     * disarm trig vendor request blocks, waiting for external trigger. This
     * has something to do with the fact we kill arm trig in siusb_cancel().
     */
    /* We always disarm trigger first.
     */
    siusb_fill_vendor_req(&s->vrs[s->nvrs++], "DISARM TRIG", USB_DIR_IN, SIUSB_REQ_DISARM_TRIG,
                          0, 0, NULL, 0,
                          SIUSB_URB_TIMEOUT,
                          NULL, 0, NULL, 0);

    BUG_ON(s->nvrs > SIUSB_MAX_VRS);
}

static void siusb_capture_start(struct siusb *s)
{
    struct siusb_buf *buf = list_entry(s->captq.next, struct siusb_buf, captq);

    s->nr_req = 0;

    /* TODO: When trying to capture small images and the image size is not an
     * exact number of packets (i.e. size mod 512 != 0), sometimes the cam does
     * not send me the last data packet. This happens constantly and only with
     * some fixed sizes. For example -- I can't transfer a 160x120 pix frame,
     * but 162x120 works perfect. Are there any restrictions on image
     * width/height/size?
     *
     * > We have seen some modulo512 issues too. Its safe to set the image size
     * > to modulo 512. We set our driver to modulo 512 even if not exact
     * > amount of data.
     */

    /* Packet-size aligned remaining bytes.
     */
    s->rem = (buf->w*buf->h+SIUSB_PACKET_SIZE-1) & ~(SIUSB_PACKET_SIZE-1);

    s->cpage = 0;

    /* Before starting actual capture, we first check if current hardware
     * settings are the same as the requested for this frame. If they aren't,
     * we change hardware and start capture after that (with the new, requested
     * settings).
     */

    /* TODO:
     *
     * Seems like changing clock, shutter mode and exposure does not work every
     * time. Is the register change order important?
     *
     * ADD: This is a bug/misfeature in camera firmware, it ACKs register
     * changes before actual change occures. It should disappear, whenever cam
     * firmware is fixed.
     */

    siusb_fill_vrs(s, buf);

    buf->result = siusb_submit_next_vr(s, buf);
    if (buf->result < 0)
        siusb_buf_complete(s, buf);
}

static void siusb_capture_request(struct siusb *s, struct siusb_buf *buf)
{
    unsigned long flags;
    int start_capture = 0;

    BUG_ON(buf->state != SIUSB_BUF_UNUSED);
    buf->state = SIUSB_BUF_QUEUED;

    buf->trig_mode = s->trig_mode;

    /* Fill frame settings from requested settings.
     */
    buf->x = s->x;
    buf->y = s->y;
    buf->w = s->w;
    buf->h = s->h;
    buf->clock = s->clock;
    buf->shutter_mode = s->shutter_mode;
    buf->exposure = s->exposure;
    buf->gain = s->gain;
    buf->black_level = s->black_level;
    buf->black_balance = s->black_balance;

    spin_lock_irqsave(&s->captq_lock, flags);

    /* If there's no readout currently running (i.e. capture queue was empty
     * until now), then we must start one.
     */
    if (list_empty(&s->captq)) {
        TRACE2("siusb: %s: buf#%u: going to start capture\n", __func__, buf->nr);
        start_capture = 1;
    }

    list_add_tail(&buf->captq, &s->captq);  /* Add to capture queue */

    spin_unlock_irqrestore(&s->captq_lock, flags);

    if (start_capture)
        siusb_capture_start(s);
}

static int siusb_readout_wait(struct siusb *s, struct siusb_buf *buf, int interruptible)
{
    int result;

    if (buf->state == SIUSB_BUF_UNUSED)
        return -EINVAL;

    TRACE1("siusb: buf#%u: %d (%s): going to sleep\n", buf->nr, current->pid, current->comm);
    if (interruptible) {
        result = wait_event_interruptible(buf->done, buf->state == SIUSB_BUF_DONE);
        if (result < 0) {
            /* Interrupted by signal.
             */
            TRACE1("siusb: buf#%u: %d (%s): awoken, interrupted by signal\n", buf->nr,
                   current->pid, current->comm);
            return result;
        }
    } else
        wait_event(buf->done, buf->state == SIUSB_BUF_DONE);
    TRACE1("siusb: buf#%u: %d (%s): awoken\n", buf->nr, current->pid, current->comm);

    buf->state = SIUSB_BUF_UNUSED;
    return 0;
}

static void siusb_cancel(struct siusb *s)
{
    struct siusb_buf *buf, *tmp_buf, *active_buf;
    unsigned long flags;
#ifdef DEBUG
    int result;
#endif

    spin_lock_irqsave(&s->captq_lock, flags);

    if (list_empty(&s->captq)) {
        /* Nothing to kill.
         */
        spin_unlock_irqrestore(&s->captq_lock, flags);
        return;
    }

    active_buf = list_entry(s->captq.next, struct siusb_buf, captq);

    /* Unlink all buffers, except the one on head. We're currently grabbing
     * into it, siusb_buf_complete() will unlink it.
     */
    list_for_each_entry_safe(buf, tmp_buf, &s->captq, captq) {
        if (buf != active_buf)
            list_del(&buf->captq);
    }

    spin_unlock_irqrestore(&s->captq_lock, flags);

    /* Now we have only one buffer in capture queue -- the one we're currently
     * grabbing into. Kill it's init URB. If the frame transfer is currently in init
     * phase (blocked waiting for external trigger), this will cause it to complete
     * (with an error) and siusb_buf_complete() will unlink this last buffer
     * from queue.
     */

    /* TODO: Should we unlink init URB only if submitted? usb_unlink_urb()
     * checks that the URB is already submitted and returns with an error if it
     * isn't.
     */
#ifdef DEBUG
    result =
#endif
    usb_unlink_urb(s->init_urb);
#ifdef DEBUG
    if (result < 0)
        TRACE1("siusb: usb_unlink_urb(init_urb)=%d: FAILED\n", result);
#endif

    /* The frame transfer is surely not in init phase now. If it was in init
     * phase, then it will complete with an error, without even starting bulk
     * transfers (because we killed the init URB). But if it wasn't in init
     * phase, then it could be currently running bulk transfers, we'll wait for
     * bulks to complete. We're not waiting for an external event here, so we
     * may safely wait uninterruptible (the bulks will complete -- sooner or
     * later).
     */
    BUG_ON(siusb_readout_wait(s, active_buf, 0) < 0);
}


/*
 * Resource allocation
 */

static void siusb_free_urbs(struct siusb *s)
{
    unsigned i;

    if (s->bulk_urbs) {
        for (i = 0; i < num_bulk_urbs; i++) {
            if (s->bulk_urbs[i])
                usb_free_urb(s->bulk_urbs[i]);
        }

        kfree(s->bulk_urbs);
        s->bulk_urbs = NULL;
    }

    if (s->init_urb) {
        usb_free_urb(s->init_urb);
        s->init_urb = NULL;
    }
}

static int siusb_alloc_urbs(struct siusb *s)
{
    unsigned i;

    /* Allocate and initialize init URB.
     */
    s->init_urb = siusb_alloc_urb(0, GFP_KERNEL);
    if (!s->init_urb) {
        printk(KERN_ERR "siusb: usb_alloc_urb(init_urb): FAILED\n");
        goto err;
    }

    s->init_urb->transfer_flags = URB_SHORT_NOT_OK;

    /* Allocate and initialize bulk URBs.
     */

    s->bulk_urbs = kmalloc(num_bulk_urbs * sizeof(*s->bulk_urbs), GFP_KERNEL);
    if (!s->bulk_urbs) {
        printk(KERN_ERR "siusb: kmalloc(bulk_urbs): FAILED\n");
        goto err;
    }

    memset(s->bulk_urbs, 0, num_bulk_urbs * sizeof(*s->bulk_urbs));

    for (i = 0; i < num_bulk_urbs; i++) {
        s->bulk_urbs[i] = siusb_alloc_urb(0, GFP_KERNEL);
        if (!s->bulk_urbs[i]) {
            printk(KERN_ERR "siusb: usb_alloc_urb(bulk_urbs[%u]): FAILED\n", i);
            goto err;
        }

        s->bulk_urbs[i]->transfer_flags = URB_SHORT_NOT_OK;
        s->bulk_urbs[i]->timeout = SIUSB_URB_TIMEOUT;
        usb_fill_bulk_urb(s->bulk_urbs[i], s->udev, usb_rcvbulkpipe(s->udev, s->bulk_ep_addr),
                          NULL, 0, siusb_bulk_urb_cb, s);
    }

    return 0;

err:
    siusb_free_urbs(s);
    return -ENOMEM;
}

static void siusb_free_buf(struct siusb_buf *buf)
{
    unsigned j;

    if (buf->pages) {
        for (j = 0; j < SIUSB_PAGES_PER_BUF; j++) {
            if (buf->pages[j])
                free_page(buf->pages[j]);
        }

        kfree(buf->pages);
        buf->pages = NULL;
    }
}

static int siusb_alloc_buf(struct siusb_buf *buf, unsigned nr)
{
    unsigned j;

    buf->nr = nr;
    buf->state = SIUSB_BUF_UNUSED;
    init_waitqueue_head(&buf->done);

    buf->pages = kmalloc(SIUSB_PAGES_PER_BUF * sizeof(*buf->pages), GFP_KERNEL);
    if (!buf->pages) {
        printk(KERN_ERR "siusb: buf#%u: kmalloc(pages): FAILED\n", buf->nr);
        goto err;
    }

    memset(buf->pages, 0, SIUSB_PAGES_PER_BUF * sizeof(*buf->pages));

    for (j = 0; j < SIUSB_PAGES_PER_BUF; j++) {
        buf->pages[j] = __get_free_page(GFP_KERNEL);
        if (!buf->pages[j]) {
            printk(KERN_ERR "siusb: buf#%u: __get_free_page(pages[%u]): FAILED\n", buf->nr, j);
            goto err;
        }
        memset((void *)buf->pages[j], 0x66, PAGE_SIZE); /* No junk to user */
    }

    return 0;

err:
    siusb_free_buf(buf);
    return -ENOMEM;
}

static void siusb_free_read_buf(struct siusb *s)
{
    siusb_free_buf(&s->read_buf);
}

static int siusb_alloc_read_buf(struct siusb *s)
{
    return siusb_alloc_buf(&s->read_buf, num_bufs);
}

static void siusb_free_bufs(struct siusb *s)
{
    unsigned i;

    if (s->bufs) {
        for (i = 0; i < num_bufs; i++)
            siusb_free_buf(&s->bufs[i]);

        kfree(s->bufs);
        s->bufs = NULL;

        TRACE1("siusb: buffers released\n");
    }
}

static int siusb_alloc_bufs(struct siusb *s)
{
    unsigned i;
    int result;

    s->bufs = kmalloc(num_bufs * sizeof(*s->bufs), GFP_KERNEL);
    if (!s->bufs) {
        printk(KERN_ERR "siusb: kmalloc(bufs): FAILED\n");
        result = -ENOMEM;
        goto err;
    }

    memset(s->bufs, 0, num_bufs * sizeof(*s->bufs));

    for (i = 0; i < num_bufs; i++) {
        result = siusb_alloc_buf(&s->bufs[i], i);
        if (result < 0)
            goto err;
    }

    TRACE1("siusb: buffers allocated\n");
    return 0;

err:
    siusb_free_bufs(s);
    return result;
}


/*
 * V4L interface
 */

static int siusb_set_size(struct siusb *s, unsigned x, unsigned y, unsigned w, unsigned h)
{
    x = (x/2)*2;
    w = (w/2)*2;

    if (w < SIUSB_MIN_WIDTH)
        return -EINVAL;
    if (w > SIUSB_MAX_WIDTH)
        w = SIUSB_MAX_WIDTH;

    if (h < SIUSB_MIN_HEIGHT)
        return -EINVAL;
    if (h > SIUSB_MAX_HEIGHT)
        h = SIUSB_MAX_HEIGHT;

    if (x + w > SIUSB_MAX_WIDTH || y + h > SIUSB_MAX_HEIGHT)
        return -EINVAL;

    s->x = x;
    s->y = y;
    s->w = w;
    s->h = h;
    return 0;
}

static int siusb_open(struct inode *inode, struct file *file)
{
    struct video_device *vdev;
    struct siusb *s;
    int result;

    if (down_interruptible(&disconnect_sem))    /* Prevent disconnects */
        return -ERESTARTSYS;

    vdev = video_devdata(file);
    if (!vdev) {
        result = -ENODEV;
        goto out0;
    }

    s = video_get_drvdata(vdev);

    if (down_interruptible(&s->sem)) {
        result = -ERESTARTSYS;
        goto out0;
    }

    /* TODO: Allow open() for using ioctls, i.e. changing exposure, gain, etc.
     */

    /* Allow single open.
     */
    if (s->open_count > 0) {
        TRACE1("siusb: %s: open_count > 0\n", __func__);
        result = -EBUSY;
        goto out;
    }

    /* First open().
     */
    result = siusb_alloc_urbs(s);       /* Allocate URBs */
    if (result < 0)
        goto out;
    result = siusb_alloc_read_buf(s);   /* Allocate read() syscall buffer */
    if (result < 0) {
        siusb_free_urbs(s);
        goto out;
    }

    s->open_count++;
    file->private_data = s;

    /* Start read() syscall streaming.
     */
    s->switched_to_mmap = 0;
    siusb_capture_request(s, &s->read_buf);

out:
    up(&s->sem);
out0:
    up(&disconnect_sem);
    return result;
}

static int siusb_release(struct inode *inode, struct file *file)
{
    struct siusb *s = file->private_data;

    /* It isn't safe to sleep interruptible here -- if we're interrupted and
     * the release() isn't restarted, open_count will not be updated properly.
     */
    down(&s->sem);

    siusb_cancel(s);

    if (--s->open_count <= 0) {     /* Last release() */
        siusb_free_bufs(s);
        siusb_free_read_buf(s);
        siusb_free_urbs(s);

        if (!s->vdev) {
            /* The camera was unplugged before the file was released.
             */
            up(&s->sem);
            TRACE1("siusb: %s: kfree(s)\n", __func__);
            kfree(s);
            return 0;
        }
    }

    up(&s->sem);
    return 0;
}

static ssize_t siusb_read(struct file *file, char *ptr, size_t count, loff_t *ppos)
{
    struct siusb *s = file->private_data;
    int rem;
    unsigned cpage;
    int result;

    if (down_interruptible(&s->sem))
        return -ERESTARTSYS;

    /* Verify that the camera wasn't unplugged.
     */
    if (!s->vdev) {
        result = -ENODEV;
        goto out;
    }

    /* Once the user has switched away from using the read() syscall, he can't
     * return back to using it without reopen.
     */
    if (s->switched_to_mmap) {
        result = -EINVAL;
        goto out;
    }

    result = siusb_readout_wait(s, &s->read_buf, 1);
    if (result < 0) {
        /* Interrupted by signal, i.e. buffer not synced yet, so we shouldn't
         * start new capture.
         */
        goto out;
    }

    if (s->read_buf.result < 0) {
        result = s->read_buf.result;
        goto out_capture_request;
    }

    if (count > s->read_buf.w * s->read_buf.h)
        count = s->read_buf.w * s->read_buf.h;

    for (rem = count, cpage = 0; rem > 0; rem -= PAGE_SIZE, cpage++) {
        if (copy_to_user(ptr, (void *)s->read_buf.pages[cpage], min(rem, (int)PAGE_SIZE))) {
            result = -EFAULT;
            goto out_capture_request;
        }

        ptr += PAGE_SIZE;
    }

    result = count;

out_capture_request:
    siusb_capture_request(s, &s->read_buf);
out:
    up(&s->sem);
    return result;
}

static unsigned siusb_poll_buf(struct file *file, struct poll_table_struct *wait,
                               struct siusb_buf *buf)
{
    if (buf->state != SIUSB_BUF_UNUSED) {
        poll_wait(file, &buf->done, wait);
        if (buf->state == SIUSB_BUF_DONE) {
            if (buf->result < 0)
                return POLLERR;
            return POLLIN | POLLRDNORM;
        }
    }
    return 0;
}

static unsigned siusb_poll(struct file *file, struct poll_table_struct *wait)
{
    struct siusb *s = file->private_data;
    unsigned i;
    unsigned mask = 0;

    /* TODO:
     *
     * Think about races here:
     * - with siusb_buf_complete() in interrupt context: removing buffers from
     *   capture queue and chaning buffer state.
     * - with syscalls queueing new buffers from user context and disconnect()
     *   changing disconnect flag.
     *
     * Should we first add ALL wait queues and after that check ready
     * conditions? Can we call poll_wait() with a spin lock held? Look at
     * poll/select internals.
     */

    /* Verify that the camera wasn't unplugged.
     */
    if (!s->vdev) {
        mask = POLLERR;
        goto out;
    }

    if (list_empty(&s->captq)) {
        /* Nothing to poll.
         */
        mask = POLLERR;
        goto out;
    }

    if (s->switched_to_mmap) {
        for (i = 0; i < num_bufs; i++)
            mask |= siusb_poll_buf(file, wait, &s->bufs[i]);
    } else
        mask = siusb_poll_buf(file, wait, &s->read_buf);

out:
    return mask;
}

static struct page *siusb_vma_nopage(struct vm_area_struct *vma, unsigned long address,
#ifdef LINUX_24
int unused
#else
int *type
#endif
)
{
    struct siusb *s = vma->vm_private_data;
    unsigned page_offset;   /* Faulting page number in this VMA */
    unsigned buf_nr;        /* Buffer number, the page belongs to */
    unsigned page_nr;       /* Page number in that buffer */
    void *pageptr;
    struct page *page;

    page_offset = vma->vm_pgoff + (address - vma->vm_start) / PAGE_SIZE;

    if (page_offset >= num_bufs * SIUSB_PAGES_PER_BUF)  /* Out of range */
        return NOPAGE_SIGBUS;

    buf_nr = page_offset / SIUSB_PAGES_PER_BUF;
    page_nr = page_offset % SIUSB_PAGES_PER_BUF;

    pageptr = (void *)s->bufs[buf_nr].pages[page_nr];
    page = virt_to_page(pageptr);

    get_page(page);
#ifndef LINUX_24
    /* 2.6.1 added this, see LWN.net Weekly Edition for January 8, 2004.
     */
    if (type)
        *type = VM_FAULT_MINOR;
#endif
    return page;
}

static struct vm_operations_struct siusb_vm_ops = {
    .nopage =   siusb_vma_nopage,
};

static int siusb_mmap(struct file *file, struct vm_area_struct *vma)
{
    struct siusb *s = file->private_data;
    int result = 0;

    if (down_interruptible(&s->sem))
        return -ERESTARTSYS;

    /* Verify that the camera wasn't unplugged.
     */
    if (!s->vdev) {
        result = -ENODEV;
        goto out;
    }

    /* Don't allow mmap() before buffers have been requested (through
     * VIDIOCGMBUF ioctl).
     */
    if (!s->switched_to_mmap) {
        result = -EINVAL;
        goto out;
    }

    /* Don't do anything here, nopage() will fill the holes.
     */
    vma->vm_flags |= VM_RESERVED;
    vma->vm_ops = &siusb_vm_ops;
    vma->vm_private_data = s;

out:
    up(&s->sem);
    return result;
}

static int siusb_do_ioctl(struct inode *inode, struct file *file, unsigned cmd, void *arg)
{
    struct siusb *s = file->private_data;
    int result = 0;

    if (down_interruptible(&s->sem))
        return -ERESTARTSYS;

    /* Verify that the camera wasn't unplugged.
     */
    if (!s->vdev) {
        result = -ENODEV;
        goto out;
    }

    switch (cmd) {
        case VIDIOCGCAP:
        {
            struct video_capability *cap = arg;

            strcpy(cap->name, s->version);
            cap->type = VID_TYPE_CAPTURE | VID_TYPE_SUBCAPTURE | VID_TYPE_MONOCHROME;
            cap->channels = 1;
            cap->audios = 0;
            cap->maxwidth = SIUSB_MAX_WIDTH;
            cap->maxheight = SIUSB_MAX_HEIGHT;
            cap->minwidth = SIUSB_MIN_WIDTH;
            cap->minheight = SIUSB_MIN_HEIGHT;
            break;
        }

        case VIDIOCGCHAN:
        {
            struct video_channel *chan = arg;

            if (chan->channel != 0) {
                result = -EINVAL;
                break;
            }
            strcpy(chan->name, "Camera");
            chan->tuners = 0;
            chan->flags = 0;
            chan->type = VIDEO_TYPE_CAMERA;
            chan->norm = VIDEO_MODE_AUTO;
            break;
        }

        case VIDIOCSCHAN:
        {
            struct video_channel *chan = arg;

            if (chan->channel != 0 || chan->norm != VIDEO_MODE_AUTO)
                result = -EINVAL;
            break;
        }

        case VIDIOCGPICT:
        {
            struct video_picture *pict = arg;

            /* TODO: More V4L settings support? Whiteness? Is that gain?
             */
            pict->brightness = s->exposure << 4;
            pict->hue = 32768;
            pict->colour = 32768;
            pict->contrast = 32768;
            pict->whiteness = 32768;
            pict->depth = 8;
            pict->palette = VIDEO_PALETTE_GREY;
            break;
        }

        case VIDIOCSPICT:
        {
            struct video_picture *pict = arg;

            if (pict->depth != 8 || pict->palette != VIDEO_PALETTE_GREY) {
                result = -EINVAL;
                break;
            }
            s->exposure = pict->brightness >> 4;
            break;
        }

        case VIDIOCGWIN:
        {
            struct video_window *win = arg;

            win->x = 0;
            win->y = 0;
            win->width = s->w;
            win->height = s->h;
            win->chromakey = 0;
            win->flags = 0;
            win->clips = NULL;
            win->clipcount = 0;
            break;
        }

        case VIDIOCSWIN:
        {
            struct video_window *win = arg;

            if (win->clips != NULL || win->clipcount != 0) {
                result = -EINVAL;
                break;
            }

            result = siusb_set_size(s, (SIUSB_MAX_WIDTH - win->width) / 2,
                                    (SIUSB_MAX_HEIGHT - win->height) / 2,
                                    win->width, win->height);
            break;
        }

        case VIDIOCGCAPTURE:
        {
            struct video_capture *capt = arg;

            capt->x = s->x;
            capt->y = s->y;
            capt->width = s->w;
            capt->height = s->h;
            capt->decimation = 0;
            capt->flags = 0;
            break;
        }

        case VIDIOCSCAPTURE:
        {
            struct video_capture *capt = arg;

            if (capt->decimation != 0 || capt->flags != 0) {
                result = -EINVAL;
                break;
            }

            result = siusb_set_size(s, capt->x, capt->y, capt->width, capt->height);
            break;
        }

        case VIDIOCGMBUF:
        {
            struct video_mbuf *mbuf = arg;
            unsigned i;

            result = siusb_alloc_bufs(s);
            if (result < 0)
                break;

            /* We can't return to using read() method again, so read() buffer
             * may be released, but be sure we aren't capturing into it now.
             */
            siusb_cancel(s);
            siusb_free_read_buf(s);

            s->switched_to_mmap = 1;

            mbuf->size = num_bufs * SIUSB_MAX_SIZE;
            mbuf->frames = num_bufs;
            for (i = 0; i < num_bufs; i++)
                mbuf->offsets[i] = i * SIUSB_MAX_SIZE;
            break;
        }

        case VIDIOCMCAPTURE:
        {
            struct video_mmap *mmap = arg;

            if (!s->switched_to_mmap) {
                result = -EINVAL;
                break;
            }

            if (mmap->frame >= num_bufs) {
                TRACE1("siusb: VIDIOCMCAPTURE: bad frame=%u\n", mmap->frame);
                result = -EINVAL;
                break;
            }

            if (mmap->format != VIDEO_PALETTE_GREY) {
                TRACE1("siusb: VIDIOCMCAPTURE: bad format=%u\n", mmap->format);
                result = -EINVAL;
                break;
            }

            if (s->bufs[mmap->frame].state != SIUSB_BUF_UNUSED) {
                TRACE1("siusb: VIDIOCMCAPTURE: frame state != SIUSB_BUF_UNUSED\n");
                result = -EBUSY;
                break;
            }

            /* Set new size only if different to prevent image centering if the
             * user has already set preferred origin and size.
             */
            if (s->w != mmap->width || s->h != mmap->height) {
                result = siusb_set_size(s, (SIUSB_MAX_WIDTH - mmap->width) / 2,
                                           (SIUSB_MAX_HEIGHT - mmap->height) / 2,
                                           mmap->width, mmap->height);
                if (result < 0)
                    break;
            }

            siusb_capture_request(s, &s->bufs[mmap->frame]);
            break;
        }

        case VIDIOCSYNC:
        {
            unsigned *frame = arg;

            if (!s->switched_to_mmap) {
                result = -EINVAL;
                break;
            }

            if (*frame >= num_bufs) {
                TRACE1("siusb: VIDIOCSYNC: bad frame=%u\n", *frame);
                result = -EINVAL;
                break;
            }

            result = siusb_readout_wait(s, &s->bufs[*frame], 1);
            if (result < 0) {
                /* Interrupted by signal.
                 */
                break;
            }

            /* TODO: What should return VIDIOCSYNC if there is frame capture error?
             *
             * We return -EINVAL on bad arg
             *           -ERESTARTSYS if signaled (probably)
             * These two mean the frame is still NOT finished and the user should reinvoke
             * VIDIOCSYNC.
             *
             * The other case is the frame is finished, but with an error -- we return
             * the error code!
             *
             * So on error return how does the user find out if the VIDIOCSYNC call must
             * be reinvoked?
             */
            result = s->bufs[*frame].result;
            break;
        }

        case VIDIOCGUNIT:
        {
            struct video_unit *unit = arg;

            unit->video = s->vdev->minor;
            unit->vbi = VIDEO_NO_UNIT;
            unit->radio = VIDEO_NO_UNIT;
            unit->audio = VIDEO_NO_UNIT;
            unit->teletext = VIDEO_NO_UNIT;
            break;
        }

        /* No overlay
         */
        case VIDIOCCAPTURE:
        case VIDIOCGFBUF:
        case VIDIOCSFBUF:
        case VIDIOCKEY:

        /* No tuner
         */
        case VIDIOCGTUNER:
        case VIDIOCSTUNER:
        case VIDIOCGFREQ:
        case VIDIOCSFREQ:

        /* No audio
         */
        case VIDIOCGAUDIO:
        case VIDIOCSAUDIO:
            result = -EINVAL;
            break;

        case SIUSB_IOC_GET_CLOCK:
        {
            unsigned *p = arg;

            *p = s->clock;
            break;
        }

        case SIUSB_IOC_SET_CLOCK:
        {
            unsigned *p = arg;

            s->clock = *p;
            break;
        }

        case SIUSB_IOC_GET_SHUTTER_MODE:
        {
            unsigned *p = arg;

            *p = s->shutter_mode;
            break;
        }

        case SIUSB_IOC_SET_SHUTTER_MODE:
        {
            unsigned *p = arg;

            s->shutter_mode = *p;
            break;
        }

        case SIUSB_IOC_GET_EXPOSURE:
        {
            unsigned *p = arg;

            *p = s->exposure;
            break;
        }

        case SIUSB_IOC_SET_EXPOSURE:
        {
            unsigned *p = arg;

            /* TODO: Shutter mode = TRIG_TB3 -- the first image line is bad.
             * This seems not to happen on TB4.
             *
             * ADD: It seems like first image line is *always* bad, I just
             * haven't seen it on TB4.
             *
             * ADD: Seems like the first line bad problem doesn't appear when
             * using external trigger mode, triggers comming every 170 msec.
             */
            /* TODO: Think again about more careful range checking for various
             * settings.
             */

            if (*p > SIUSB_EXPOSURE_MAX) {
                result = -EINVAL;
                break;
            }
            s->exposure = *p;
            break;
        }

        case SIUSB_IOC_GET_GAIN:
        {
            unsigned *p = arg;

            *p = s->gain;
            break;
        }

        case SIUSB_IOC_SET_GAIN:
        {
            unsigned *p = arg;

            if (*p < SIUSB_GAIN_MIN || *p > SIUSB_GAIN_UNITY) {
                result = -EINVAL;
                break;
            }
            s->gain = *p;
            break;
        }

        case SIUSB_IOC_GET_BLACK_LEVEL:
        {
            unsigned *p = arg;

            *p = s->black_level;
            break;
        }

        case SIUSB_IOC_SET_BLACK_LEVEL:
        {
            unsigned *p = arg;

            if (*p > SIUSB_BLACK_LEVEL_MAX) {
                result = -EINVAL;
                break;
            }
            s->black_level = *p;
            break;
        }

        case SIUSB_IOC_GET_BLACK_BALANCE:
        {
            unsigned *p = arg;

            *p = s->black_balance;
            break;
        }

        case SIUSB_IOC_SET_BLACK_BALANCE:
        {
            unsigned *p = arg;

            if (*p > SIUSB_BLACK_BALANCE_MAX) {
                result = -EINVAL;
                break;
            }
            s->black_balance = *p;
            break;
        }

        case SIUSB_IOC_GET_TRIG_MODE:
        {
            unsigned *p = arg;

            *p = s->trig_mode;
            break;
        }

        case SIUSB_IOC_SET_TRIG_MODE:
        {
            unsigned *p = arg;

            s->trig_mode = *p;
            break;
        }

        default:
            result = -ENOIOCTLCMD;
            break;
    }

out:
    up(&s->sem);
    return result;
}

static int siusb_ioctl(struct inode *inode, struct file *file, unsigned cmd, unsigned long arg)
{
    return video_usercopy(inode, file, cmd, arg, siusb_do_ioctl);
}

static struct file_operations siusb_fops = {
    .owner =    THIS_MODULE,
    .open =     siusb_open,
    .release =  siusb_release,
    .read =     siusb_read,
    .poll =     siusb_poll,
    .mmap =     siusb_mmap,
    .ioctl =    siusb_ioctl,
    .llseek =   no_llseek,
};

static struct video_device siusb_vdev_template = {
    .owner =    THIS_MODULE,
    .type =     VID_TYPE_CAPTURE | VID_TYPE_SUBCAPTURE | VID_TYPE_MONOCHROME,
    .hardware = VID_HARDWARE_SI,
    .fops =     &siusb_fops,
    .release =  video_device_release,
};

static int siusb_v4l_register(struct siusb *s)
{
    int result;

    s->vdev = video_device_alloc();
    if (!s->vdev) {
        printk(KERN_ERR "siusb: video_device_alloc(): FAILED\n");
        return -ENOMEM;
    }

    *s->vdev = siusb_vdev_template;
#ifndef LINUX_24
    s->vdev->dev = &s->udev->dev;
#endif
    strcpy(s->vdev->name, s->version);
    video_set_drvdata(s->vdev, s);

    /* TODO: video_nr module parameter? How to pass it if we have more than one
     * camera.
     */
    result = video_register_device(s->vdev, VFL_TYPE_GRABBER, -1);
    if (result < 0) {
        printk(KERN_ERR "siusb: video_register_device()=%d: FAILED\n", result);
        video_device_release(s->vdev);
        return result;
    }

    return 0;
}

static void siusb_v4l_unregister(struct siusb *s)
{
    video_unregister_device(s->vdev);
}


/*
 * Initialization
 */

static int siusb_init_cam(struct siusb *s)
{
    int result;

    result = siusb_set_subsampling(s, SIUSB_SUBSAMP_NONE);
    if (result < 0)
        return result;

    result = siusb_set_depth(s, SIUSB_DEPTH_8BIT);
    if (result < 0)
        return result;

    s->x = 0;
    s->y = 0;
    s->w = SIUSB_MAX_WIDTH;
    s->h = SIUSB_MAX_HEIGHT;
    s->clock = SIUSB_CLK_20MHZ;
    s->shutter_mode = SIUSB_SH_TRIG_TB3;
    s->exposure = 2000;
    s->gain = SIUSB_GAIN_1_88;
    s->black_level = 85;
    s->black_balance = 85;

    /* Sensor registers initialization is defered for the first frame capture.
     * We initialize current in-hardware settings with fake values. On the
     * first frame capture, hardware settings will be set to match the
     * requested ones.
     */
    s->hw_x = -1;
    s->hw_y = -1;
    s->hw_w = -1;
    s->hw_h = -1;
    s->hw_clock = -1;
    s->hw_shutter_mode = -1;
    s->hw_exposure = -1;
    s->hw_gain = -1;
    s->hw_black_level = -1;
    s->hw_black_balance = -1;

    return 0;
}

#ifdef LINUX_24
static void *siusb_probe(struct usb_device *udev, unsigned ifnum, const struct usb_device_id *id)
#else
static int siusb_probe(struct usb_interface *ifc, const struct usb_device_id *id)
#endif
{
    struct siusb *s;
#ifdef LINUX_24
    struct usb_interface *ifc;
#endif
    int result;
    struct usb_endpoint_descriptor *ep_desc;
    char cpld[SIUSB_CPLD_SIZE];

    /* Allocate memory for our device.
     */
    s = kmalloc(sizeof(*s), GFP_KERNEL);
    if (!s) {
        printk(KERN_ERR "siusb: kmalloc(s): FAILED\n");
        result = -ENOMEM;
        goto err;
    }

    /* And initialize it.
     */
    memset(s, 0, sizeof(*s));
#ifdef LINUX_24
    s->udev = udev;
    ifc = usb_ifnum_to_if(udev, ifnum);
#else
    s->udev = interface_to_usbdev(ifc);
#endif
    init_MUTEX(&s->sem);

    /* Init capture queue.
     */
    INIT_LIST_HEAD(&s->captq);
    spin_lock_init(&s->captq_lock);

    /* Do some sanity checks.
     */
    if (s->udev->speed != USB_SPEED_HIGH) {
        printk(KERN_ERR "siusb: high speed HC/HCD required\n");
        result = -ENODEV;
        goto err;
    }
    if (s->udev->descriptor.bNumConfigurations != 1) {
        printk(KERN_ERR "siusb: bNumConfigurations != 1\n");
        result = -ENODEV;
        goto err;
    }
#ifdef LINUX_24
    ep_desc = &ifc->altsetting[0].endpoint[0];
#else
    ep_desc = &ifc->altsetting[0].endpoint[0].desc;
#endif
    if (!(ep_desc->bEndpointAddress & USB_DIR_IN)
     || !((ep_desc->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_BULK)) {
        printk(KERN_ERR "siusb: endpoint not bulk IN\n");
        result = -ENODEV;
        goto err;
    }
    if (ep_desc->wMaxPacketSize != SIUSB_PACKET_SIZE) {
        printk(KERN_ERR "siusb: wMaxPacketSize != " __stringify(SIUSB_PACKET_SIZE) "\n");
        result = -ENODEV;
        goto err;
    }

    s->bulk_ep_addr = ep_desc->bEndpointAddress;
    TRACE2("siusb: bulk_ep_addr=0x%02x\n", s->bulk_ep_addr);

    /* Wait for camera to initialize, it may have been just plugged in.
     */
    set_current_state(TASK_UNINTERRUPTIBLE);
    schedule_timeout(HZ);   /* 1 sec */

    /* Read camera firmware version.
     */
    result = siusb_read_version(s, s->version);
    if (result < 0)
        goto err;

    /* Read camera CPLD version.
     */
    result = siusb_read_cpld(s, cpld);
    if (result < 0)
        goto err;

    /* Initialize camera registers.
     */
    result = siusb_init_cam(s);
    if (result < 0)
        goto err;

    result = siusb_v4l_register(s);
    if (result < 0)
        goto err;

    /* TODO: proc and/or sysfs interface.
     */

    printk(KERN_INFO "siusb: attached: CPLD %c%c%c%c, firmware %s, V4L minor %d\n",
           cpld[0], cpld[1], cpld[2], cpld[3], s->version, s->vdev->minor);
#ifdef LINUX_24
    return s;
#else
    usb_set_intfdata(ifc, s);
    return 0;
#endif

err:
    TRACE1("siusb: %s: kfree(s)\n", __func__);
    kfree(s);
#ifdef LINUX_24
    return NULL;
#else
    return result;
#endif
}

#ifdef LINUX_24
static void siusb_disconnect(struct usb_device *udev, void *context)
#else
static void siusb_disconnect(struct usb_interface *ifc)
#endif
{
    struct siusb *s;

    down(&disconnect_sem);      /* Prevent races with open() */

#ifdef LINUX_24
    s = context;
#else
    s = usb_get_intfdata(ifc);
    usb_set_intfdata(ifc, NULL);
#endif

    down(&s->sem);
    siusb_v4l_unregister(s);
    s->vdev = NULL;             /* Prevent file operations */
    up(&s->sem);

    /* If the device is opened, siusb_release() will clean this up.
     */
    if (s->open_count <= 0) {
        TRACE1("siusb: %s: kfree(s)\n", __func__);
        kfree(s);
    }

    up(&disconnect_sem);
    printk(KERN_INFO "siusb: disconnected\n");
}

static struct usb_device_id siusb_id_table[] = {
    {USB_DEVICE(0x0547, 0x1002)},
    { } /* Terminating entry */
};
MODULE_DEVICE_TABLE(usb, siusb_id_table);

static struct usb_driver siusb_driver = {
    .owner =        THIS_MODULE,
    .name =         "siusb",
    .probe =        siusb_probe,
    .disconnect =   siusb_disconnect,
    .id_table =     siusb_id_table,
};

static int __init siusb_init(void)
{
    int result;

    printk(KERN_INFO "siusb: " DRIVER_DESC " " DRIVER_VERSION "\n");
    result = usb_register(&siusb_driver);
    if (result < 0)
        printk(KERN_ERR "siusb: usb_register()=%d: FAILED\n", result);
    return result;
}

static void __exit siusb_exit(void)
{
    usb_deregister(&siusb_driver);
}

module_init(siusb_init);
module_exit(siusb_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("Proprietary");
