/*
 * siusbctl.c
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>

#include "sicam.h"

/*
 * Clock
 */

struct siusb_clock {
    unsigned clock;
    unsigned val;
};

#define SIUSB_NR_CLOCK_VALUES   9

static struct siusb_clock siusb_clock_values[SIUSB_NR_CLOCK_VALUES] = {
    {20, SIUSB_CLK_20MHZ},
    {25, SIUSB_CLK_25MHZ},
    {30, SIUSB_CLK_30MHZ},
    {35, SIUSB_CLK_35MHZ},
    {40, SIUSB_CLK_40MHZ},
    {45, SIUSB_CLK_45MHZ},
    {50, SIUSB_CLK_50MHZ},
    {55, SIUSB_CLK_55MHZ},
    {60, SIUSB_CLK_60MHZ}
};

static unsigned val_to_clock(unsigned val)
{
    unsigned i;

    for (i = 0; i < SIUSB_NR_CLOCK_VALUES; i++) {
        if (val == siusb_clock_values[i].val)
            return siusb_clock_values[i].clock;
    }
    fprintf(stderr, "unknown clock value\n");
    exit(1);
}

static unsigned clock_to_val(unsigned clock)
{
    unsigned i;

    for (i = 0; i < SIUSB_NR_CLOCK_VALUES; i++) {
        if (clock == siusb_clock_values[i].clock)
            return siusb_clock_values[i].val;
    }
    fprintf(stderr, "bad clock, allowed are:\n");
    for (i = 0; i < SIUSB_NR_CLOCK_VALUES; i++)
        fprintf(stderr, "%u%s", siusb_clock_values[i].clock,
                i == SIUSB_NR_CLOCK_VALUES-1 ? "\n" : ", ");
    exit(1);
}

/*
 * Shutter mode
 */

struct siusb_shutter_mode {
    char *mode;
    unsigned val;
};

#define SIUSB_NR_SHUTTER_MODE_VALUES    8

static struct siusb_shutter_mode siusb_shutter_mode_values[SIUSB_NR_SHUTTER_MODE_VALUES] = {
    {"ROLLING",     SIUSB_SH_ROLLING    },
    {"ROLLING_DS1", SIUSB_SH_ROLLING_DS1},
    {"ROLLING_DS2", SIUSB_SH_ROLLING_DS2},
    {"ROLLING_DS3", SIUSB_SH_ROLLING_DS3},
    {"TRIG_TB1",    SIUSB_SH_TRIG_TB1   },
    {"TRIG_TB2",    SIUSB_SH_TRIG_TB2   },
    {"TRIG_TB3",    SIUSB_SH_TRIG_TB3   },
    {"TRIG_TB4",    SIUSB_SH_TRIG_TB4   }
};

static char *val_to_shutter_mode(unsigned val)
{
    unsigned i;

    for (i = 0; i < SIUSB_NR_SHUTTER_MODE_VALUES; i++) {
        if (val == siusb_shutter_mode_values[i].val)
            return siusb_shutter_mode_values[i].mode;
    }
    fprintf(stderr, "unknown shutter mode value\n");
    exit(1);
}

static unsigned shutter_mode_to_val(const char *mode)
{
    unsigned i;

    for (i = 0; i < SIUSB_NR_SHUTTER_MODE_VALUES; i++) {
        if (strcasecmp(mode, siusb_shutter_mode_values[i].mode) == 0)
            return siusb_shutter_mode_values[i].val;
    }
    fprintf(stderr, "bad shutter mode, allowed are:\n");
    for (i = 0; i < SIUSB_NR_SHUTTER_MODE_VALUES; i++)
        fprintf(stderr, "%s%s", siusb_shutter_mode_values[i].mode,
                i == SIUSB_NR_SHUTTER_MODE_VALUES-1 ? "\n" : ", ");
    exit(1);
}

/*
 * Gain
 */

struct siusb_gain {
    double gain;
    unsigned val;
};

#define SIUSB_NR_GAIN_VALUES    17

static struct siusb_gain siusb_gain_values[SIUSB_NR_GAIN_VALUES] = {
    {1.41,  SIUSB_GAIN_1_41 },
    {1.62,  SIUSB_GAIN_1_62 },
    {1.88,  SIUSB_GAIN_1_88 },
    {2.17,  SIUSB_GAIN_2_17 },
    {2.52,  SIUSB_GAIN_2_52 },
    {2.92,  SIUSB_GAIN_2_92 },
    {3.38,  SIUSB_GAIN_3_38 },
    {3.90,  SIUSB_GAIN_3_90 },
    {4.50,  SIUSB_GAIN_4_50 },
    {5.17,  SIUSB_GAIN_5_17 },
    {5.93,  SIUSB_GAIN_5_93 },
    {6.76,  SIUSB_GAIN_6_76 },
    {7.86,  SIUSB_GAIN_7_86 },
    {8.64,  SIUSB_GAIN_8_64 },
    {9.61,  SIUSB_GAIN_9_61 },
    {10.69, SIUSB_GAIN_10_69},
    {1.00,  SIUSB_GAIN_1_00 },
};

static double val_to_gain(unsigned val)
{
    unsigned i;

    for (i = 0; i < SIUSB_NR_GAIN_VALUES; i++) {
        if (val == siusb_gain_values[i].val)
            return siusb_gain_values[i].gain;
    }
    fprintf(stderr, "unknown gain value\n");
    exit(1);
}

static unsigned gain_to_val(double gain)
{
    unsigned i;

    for (i = 0; i < SIUSB_NR_GAIN_VALUES; i++) {
        if (gain == siusb_gain_values[i].gain)
            return siusb_gain_values[i].val;
    }
    fprintf(stderr, "bad gain, allowed are:\n");
    for (i = 0; i < SIUSB_NR_GAIN_VALUES; i++)
        fprintf(stderr, "%.2f%s", siusb_gain_values[i].gain,
                i == SIUSB_NR_GAIN_VALUES-1 ? "\n" : ", ");
    exit(1);
}

#define CLOCK_OPT           0
#define SHUTTER_MODE_OPT    1
#define EXPOSURE_OPT        2
#define GAIN_OPT            3
#define BLACK_LEVEL_OPT     4
#define BLACK_BALANCE_OPT   5
#define TRIG_MODE_OPT       6

#define NUMOPTS             7

static int option_present[NUMOPTS];
static const char *option_str[NUMOPTS];

static const struct option long_options[NUMOPTS+1] = {
    {"clock",         optional_argument, &option_present[CLOCK_OPT],         1},
    {"shutter-mode",  optional_argument, &option_present[SHUTTER_MODE_OPT],  1},
    {"exposure",      optional_argument, &option_present[EXPOSURE_OPT],      1},
    {"gain",          optional_argument, &option_present[GAIN_OPT],          1},
    {"black-level",   optional_argument, &option_present[BLACK_LEVEL_OPT],   1},
    {"black-balance", optional_argument, &option_present[BLACK_BALANCE_OPT], 1},
    {"trig-mode",     optional_argument, &option_present[TRIG_MODE_OPT],     1},
    {NULL,            0,                 NULL,                               0}
};

static void set_clock(struct cam *c)
{
    if (sicam_set_clock(c, clock_to_val(atoi(option_str[CLOCK_OPT]))))
        exit(1);
}

static void get_clock(struct cam *c)
{
    printf("clock=%u\n", val_to_clock(sicam_get_clock(c)));
}

static void set_shutter_mode(struct cam *c)
{
    if (sicam_set_shutter_mode(c, shutter_mode_to_val(option_str[SHUTTER_MODE_OPT])))
        exit(1);
}

static void get_shutter_mode(struct cam *c)
{
    printf("shutter-mode=%s\n", val_to_shutter_mode(sicam_get_shutter_mode(c)));
}

static void set_exposure(struct cam *c)
{
    if (sicam_set_exposure(c, atoi(option_str[EXPOSURE_OPT])))
        exit(1);
}

static void get_exposure(struct cam *c)
{
    printf("exposure=%u\n", sicam_get_exposure(c));
}

static void set_gain(struct cam *c)
{
    if (sicam_set_gain(c, gain_to_val(atof(option_str[GAIN_OPT]))))
        exit(1);
}

static void get_gain(struct cam *c)
{
    printf("gain=%.2f\n", val_to_gain(sicam_get_gain(c)));
}

static void set_black_level(struct cam *c)
{
    if (sicam_set_black_level(c, atoi(option_str[BLACK_LEVEL_OPT])))
        exit(1);
}

static void get_black_level(struct cam *c)
{
    printf("black-level=%u\n", sicam_get_black_level(c));
}

static void set_black_balance(struct cam *c)
{
    if (sicam_set_black_balance(c, atoi(option_str[BLACK_BALANCE_OPT])))
        exit(1);
}

static void get_black_balance(struct cam *c)
{
    printf("black-balance=%u\n", sicam_get_black_balance(c));
}

static void set_trig_mode(struct cam *c)
{
    sicam_set_trig_mode(c, atoi(option_str[TRIG_MODE_OPT]));
}

static void get_trig_mode(struct cam *c)
{
    printf("trig-mode=%u\n", sicam_get_trig_mode(c));
}

static void usage(const char *argv0)
{
    fprintf(stderr, "\n");
    fprintf(stderr, "Usage: %s [OPTIONS] VIDEO_DEVICE\n", argv0);
    fprintf(stderr, "\n");
    fprintf(stderr, "Options:\n");
    fprintf(stderr, " -c, --clock[=CLOCK]                   get/set clock\n");
    fprintf(stderr, " -s, --shutter-mode[=SHUTTER_MODE]     get/set shutter mode\n");
    fprintf(stderr, " -e, --exposure[=EXPOSURE]             get/set exposure\n");
    fprintf(stderr, " -g, --gain[=GAIN]                     get/set gain\n");
    fprintf(stderr, " -l, --black-level[=BLACK_LEVEL]       get/set black level\n");
    fprintf(stderr, " -b, --black-balance[=BLACK_BALANCE]   get/set black balance\n");
    fprintf(stderr, " -t, --trig-mode[=TRIG_MODE]           get/set trigger mode\n");
}

int main(int argc, char *argv[])
{
    int ch;
    int option_index;
    char *dev;
    struct cam *c;

    while (1) {
        ch = getopt_long(argc, argv, "c::s::e::g::l::b::t::", long_options, &option_index);

        if (ch == -1)
            break;

        switch (ch) {
            case 0:     option_str[option_index] = optarg;
                        break;

            case 'c':   option_present[CLOCK_OPT] = 1;
                        option_str[CLOCK_OPT] = optarg;
                        break;

            case 's':   option_present[SHUTTER_MODE_OPT] = 1;
                        option_str[SHUTTER_MODE_OPT] = optarg;
                        break;

            case 'e':   option_present[EXPOSURE_OPT] = 1;
                        option_str[EXPOSURE_OPT] = optarg;
                        break;

            case 'g':   option_present[GAIN_OPT] = 1;
                        option_str[GAIN_OPT] = optarg;
                        break;

            case 'l':   option_present[BLACK_LEVEL_OPT] = 1;
                        option_str[BLACK_LEVEL_OPT] = optarg;
                        break;

            case 'b':   option_present[BLACK_BALANCE_OPT] = 1;
                        option_str[BLACK_BALANCE_OPT] = optarg;
                        break;

            case 't':   option_present[TRIG_MODE_OPT] = 1;
                        option_str[TRIG_MODE_OPT] = optarg;
                        break;

            case '?':   return 1;   /* getopt_long() already printed an error message */
            default:    abort();    /* Should never happen */
        }
    }

    if (optind < argc)
        dev = argv[optind];
    else {
        usage(argv[0]);
        return 1;
    }

    c = cam_open(dev);
    if (!c)
        exit(1);

    if (option_present[CLOCK_OPT]) {
        if (option_str[CLOCK_OPT])
            set_clock(c);
        else
            get_clock(c);
    }

    if (option_present[SHUTTER_MODE_OPT]) {
        if (option_str[SHUTTER_MODE_OPT])
            set_shutter_mode(c);
        else
            get_shutter_mode(c);
    }

    if (option_present[EXPOSURE_OPT]) {
        if (option_str[EXPOSURE_OPT])
            set_exposure(c);
        else
            get_exposure(c);
    }

    if (option_present[GAIN_OPT]) {
        if (option_str[GAIN_OPT])
            set_gain(c);
        else
            get_gain(c);
    }

    if (option_present[BLACK_LEVEL_OPT]) {
        if (option_str[BLACK_LEVEL_OPT])
            set_black_level(c);
        else
            get_black_level(c);
    }

    if (option_present[BLACK_BALANCE_OPT]) {
        if (option_str[BLACK_BALANCE_OPT])
            set_black_balance(c);
        else
            get_black_balance(c);
    }

    if (option_present[TRIG_MODE_OPT]) {
        if (option_str[TRIG_MODE_OPT])
            set_trig_mode(c);
        else
            get_trig_mode(c);
    }

    /* If no options specified, print all.
     */

    for (option_index = 0; option_index < NUMOPTS; option_index++) {
        if (option_present[option_index])
            break;
    }

    if (option_index >= NUMOPTS) {
        get_clock(c);
        get_shutter_mode(c);
        get_exposure(c);
        get_gain(c);
        get_black_level(c);
        get_black_balance(c);
        get_trig_mode(c);
    }

    cam_close(c);
    return 0;
}
