/*
 * siusbemu.c
 *
 * SI-1280F USB Camera Driver Emulator
 *
 * Copyright (C) 2003-2004 Nikka Research Deutschland GmbH
 */

/*
 * TODO:
 *
 *   Set clock transfer size must be 3 bytes, check that? How?
 *
 *   Allow open() for using ioctl's(), i.e. changing exposure, gain, etc.
 *
 *   proc and/or sysfs interface.
 *
 *   Careful range checking for various settings (exposure, etc.).
 *
 *   More V4L settings support? Whiteness? Is that gain?
 *
 *   Seems like changing exposure does not work every time. Is the order of register setting
 *   important?
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <linux/videodev.h>

#include "siusb.h"


/* Kernel compatibility
 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#define LINUX_24
#endif


/* Debug macros
 */
#ifdef DEBUG
#define TRACE1(fmt, args...)    do { if (DEBUG >= 1) printk(KERN_DEBUG fmt, ## args); } while (0)
#define TRACE2(fmt, args...)    do { if (DEBUG >= 2) printk(KERN_DEBUG fmt, ## args); } while (0)
#define TRACE3(fmt, args...)    do { if (DEBUG >= 3) printk(KERN_DEBUG fmt, ## args); } while (0)
#else /* !DEBUG */
#define TRACE1(fmt, args...)    do { } while (0)
#define TRACE2(fmt, args...)    do { } while (0)
#define TRACE3(fmt, args...)    do { } while (0)
#endif /* !DEBUG */


#define DRIVER_AUTHOR   "Nikka Research Deutschland GmbH"
#define DRIVER_DESC     "SI-1280F USB Camera Driver Emulator"
#define DRIVER_VERSION  "v0.3 based on siusb v0.7.2"


/* Module parameters
 */

#ifndef SIUSB_NUM_BUFS
#define SIUSB_NUM_BUFS      2
#endif

static unsigned num_bufs = SIUSB_NUM_BUFS;

MODULE_PARM(num_bufs, "i");
MODULE_PARM_DESC(num_bufs,
    "Number of buffers (default is " __stringify(SIUSB_NUM_BUFS) ")");

#ifndef SIUSB_TRIG_INTERVAL
#define SIUSB_TRIG_INTERVAL 500
#endif

static unsigned trig_interval = SIUSB_TRIG_INTERVAL;

MODULE_PARM(trig_interval, "i");
MODULE_PARM_DESC(trig_interval,
    "Trigger interval [ms] (default is " __stringify(SIUSB_TRIG_INTERVAL) ")");


#define SIUSB_PACKET_SIZE   512

#define SIUSB_VERSION_SIZE  16


#define SIUSB_MIN_WIDTH     256
#define SIUSB_MIN_HEIGHT    1

#define SIUSB_MAX_WIDTH     1600
#define SIUSB_MAX_HEIGHT    1216

#define SIUSB_MAX_SIZE      (SIUSB_MAX_WIDTH * SIUSB_MAX_HEIGHT)


#define SIUSB_INIT_URB_TIMEOUT  (HZ/2)
#define SIUSB_BULK_URB_TIMEOUT  (HZ/2)

#define SIUSB_CTRL_TIMEOUT      (5*HZ)  /* Control transfer timeout in jiffies */


/* Origin, width, height, clock, shutter mode, exposure, gain, black level,
 * black balance -- total 9.
 */
#define SIUSB_MAX_VRS       9


#ifndef VID_HARDWARE_SI
#define VID_HARDWARE_SI     71232   /* Random */
#endif


/* Buffer states
 */
#define SIUSB_BUF_UNUSED    0   /* Unused (no MCAPTURE) */
#define SIUSB_BUF_QUEUED    1   /* In capture queue */
#define SIUSB_BUF_DONE      2   /* Finished grabbing, but not been synced yet */

struct siusb_buf {
    unsigned nr;                /* Index in buffers array */
    unsigned state;             /* Buffer state */
    int result;                 /* Negative error code or 0 = no error */

    unsigned trig_mode;         /* Internal or external trigger mode */

    /* Requested register settings for this frame
     */
    unsigned x, y, w, h;
    unsigned clock;
    unsigned shutter_mode;
    unsigned exposure;
    unsigned gain;
    unsigned black_level;
    unsigned black_balance;

    unsigned long *pages;       /* Pages array */
    wait_queue_head_t done;     /* Wait for readout finish */
    struct list_head captq;     /* Capture queue list */

    struct timer_list timer;
};


struct siusb {
    struct video_device *vdev;  /* V4L device for that camera */

    int open_count;             /* Device open count */
    int present;                /* True if device is not disconnected */
    struct semaphore sem;       /* Locks this structure */
    int bulk_ep_addr;           /* Bulk IN endpoint address */

    char version[SIUSB_VERSION_SIZE];   /* Keep camera version string */

    struct siusb_buf *bufs;     /* Buffers array */
    struct siusb_buf read_buf;  /* read() syscall buffer */

    int switched_to_mmap;       /* True if switched to using mmap() instead of read() */

    struct list_head captq;     /* Capture queue */
    spinlock_t captq_lock;      /* Lock capture queue add/remove */

    int nr_req;                 /* Number of queued bulk transfer requests */
    int rem;                    /* Remaining bytes */
    unsigned cpage;             /* Current page in pages array */

    unsigned trig_mode;         /* Trigger mode */

    /* Current in-hardware register settings
     */
    unsigned hw_x, hw_y, hw_w, hw_h;
    unsigned hw_clock;
    unsigned hw_shutter_mode;
    unsigned hw_exposure;
    unsigned hw_gain;
    unsigned hw_black_level;
    unsigned hw_black_balance;

    /* Requested register settings
     */
    unsigned x, y, w, h;
    unsigned clock;
    unsigned shutter_mode;
    unsigned exposure;
    unsigned gain;
    unsigned black_level;
    unsigned black_balance;
};


static unsigned siusb_pages_per_buf;    /* Numer of pages in pages array of each buffer */

static DECLARE_MUTEX(disconnect_sem);   /* Prevent races between open() and disconnect() */


/*
 * Low-level
 */

/* Vendor request codes
 */
#define SIUSB_REQ_VERSION           0xE1    /* Read firmware version */
#define SIUSB_REQ_FLASH             0xE4    /* Read/Write flash memory */
#define SIUSB_REQ_CLOCK             0xE2    /* Get/Set clock */
#define SIUSB_REQ_LOAD_REG          0xEF    /* Load sensor registers */
#define SIUSB_REQ_SET_ORIGIN        0xE6    /* Set row/column origin (pan/tilt) */
#define SIUSB_REQ_SET_DEPTH         0xED    /* Set bit depth (8 or 16 bits) */
#define SIUSB_REQ_READOUT           0xEA    /* Snapshot expose & readout */
#define SIUSB_REQ_ARM_TRIG          0xEC    /* Arm triggered, snapshot exposure mode */
#define SIUSB_REQ_DISARM_TRIG       0xEB    /* Disarm triggered mode */

/* Registers
 */
#define SIUSB_REG_SHUTTER_MODE      0x00
#define SIUSB_REG_WIDTH             0x01    /* (Pixels in row) / 2 - 1 */
#define SIUSB_REG_HEIGHT            0x02    /* Number of rows - 1 */
#define SIUSB_REG_EXPOSURE          0x03
#define SIUSB_REG_BEGIN_COLUMN      0x04    /* X-origin / 2 */
#define SIUSB_REG_BEGIN_ROW         0x05    /* Y-origin */
#define SIUSB_REG_BEGIN_ROW_2       0x06    /* Set to same value as Reg5 */
#define SIUSB_REG_SUBSAMPLING       0x07
#define SIUSB_REG_GAIN              0x08
#define SIUSB_REG_BLACK_LEVEL       0x09
#define SIUSB_REG_BLACK_BALANCE     0x0A

/* Bit depth
 */
#define SIUSB_DEPTH_8BIT    0
#define SIUSB_DEPTH_16BIT   1


static int siusb_read_version(struct siusb *s, char *buf)
{
    strcpy(buf, "SIUSBEmu");
    return 0;
}

static int siusb_set_depth(struct siusb *s, int depth)
{
    return 0;
}


/*
 * Capture queue
 */

static void siusb_capture_start(struct siusb *s);

static void siusb_buf_complete(struct siusb *s, struct siusb_buf *buf)
{
    unsigned long flags;
    unsigned start_capture = 0;

    /* Frame transfer is complete, when all queued URBs are complete.
     */
    if (s->nr_req > 0)
        return;

    TRACE2("siusb: buf#%u: DONE, result=%d\n", buf->nr, buf->result);

    spin_lock_irqsave(&s->captq_lock, flags);

    list_del(&buf->captq);

    if (!list_empty(&s->captq)) {
        TRACE2("siusb: %s: buf#%u: going to start capture\n", __func__, buf->nr);
        start_capture = 1;
    }

    spin_unlock_irqrestore(&s->captq_lock, flags);

    buf->state = SIUSB_BUF_DONE;
    TRACE2("siusb: buf#%u: wakeing up\n", buf->nr);
    wake_up_interruptible(&buf->done);

    if (start_capture)
        siusb_capture_start(s);
}

static void siusb_do_complete(unsigned long data)
{
    struct siusb *s = (struct siusb *)data;
    struct siusb_buf *buf = list_entry(s->captq.next, struct siusb_buf, captq);

    siusb_buf_complete(s, buf);
}

static void siusb_fill_vrs(struct siusb *s, struct siusb_buf *buf)
{
    s->hw_black_balance = buf->black_balance;
    s->hw_black_level = buf->black_level;
    s->hw_gain = buf->gain;
    s->hw_exposure = buf->exposure;
    s->hw_shutter_mode = buf->shutter_mode;
    s->hw_clock = buf->clock;
    s->hw_h = buf->h;
    s->hw_w = buf->w;
    s->hw_x = buf->x;
    s->hw_y = buf->y;
}

static void siusb_capture_start(struct siusb *s)
{
    struct siusb_buf *buf = list_entry(s->captq.next, struct siusb_buf, captq);

    s->nr_req = 0;

    /* Packet-size aligned remaining bytes.
     */
    s->rem = (buf->w*buf->h+SIUSB_PACKET_SIZE-1) & ~(SIUSB_PACKET_SIZE-1);

    s->cpage = 0;

    /* Before starting actual capture, we first check if current hardware
     * settings are the same as the requested for this frame. If they
     * aren't, we change hardware and start capture after that (with the new,
     * requested settings).
     */

    siusb_fill_vrs(s, buf);

    /* Add a kernel timer to run complete after some fixed time.
     */
    init_timer(&buf->timer);
    buf->timer.expires = jiffies + trig_interval*HZ/1000;
    buf->timer.data = (unsigned long)s;
    buf->timer.function = siusb_do_complete;
    add_timer(&buf->timer);
}

static void siusb_capture_request(struct siusb *s, struct siusb_buf *buf)
{
    unsigned long flags;
    int start_capture = 0;

    BUG_ON(buf->state != SIUSB_BUF_UNUSED);
    buf->state = SIUSB_BUF_QUEUED;

    buf->trig_mode = s->trig_mode;

    /* Fill frame settings from requested settings.
     */
    buf->x = s->x;
    buf->y = s->y;
    buf->w = s->w;
    buf->h = s->h;
    buf->clock = s->clock;
    buf->shutter_mode = s->shutter_mode;
    buf->exposure = s->exposure;
    buf->gain = s->gain;
    buf->black_level = s->black_level;
    buf->black_balance = s->black_balance;

    spin_lock_irqsave(&s->captq_lock, flags);

    /* If there's no readout currently running (i.e. capture queue was empty
     * until now), then we must start one.
     */
    if (list_empty(&s->captq)) {
        TRACE2("siusb: %s: buf#%u: going to start capture\n", __func__, buf->nr);
        start_capture = 1;
    }

    list_add_tail(&buf->captq, &s->captq);

    spin_unlock_irqrestore(&s->captq_lock, flags);

    if (start_capture)
        siusb_capture_start(s);
}

static int siusb_readout_wait(struct siusb *s, struct siusb_buf *buf)
{
    int result;

    if (buf->state == SIUSB_BUF_UNUSED)
        return -EINVAL;

    TRACE2("siusb: buf#%u: %d (%s): going to sleep\n", buf->nr, current->pid, current->comm);
    result = wait_event_interruptible(buf->done, buf->state == SIUSB_BUF_DONE);
    TRACE2("siusb: buf#%u: %d (%s): awoken\n", buf->nr, current->pid, current->comm);
    if (result < 0) {
        /* Interrupted by signal.
         */
        return result;
    }

    buf->state = SIUSB_BUF_UNUSED;
    return 0;
}

static int siusb_cancel(struct siusb *s)
{
    struct siusb_buf *buf, *tmp_buf, *active_buf;
    unsigned long flags;
    int result;

    spin_lock_irqsave(&s->captq_lock, flags);

    if (list_empty(&s->captq)) {
        /* Nothing to kill.
         */
        spin_unlock_irqrestore(&s->captq_lock, flags);
        return 0;
    }

    active_buf = list_entry(s->captq.next, struct siusb_buf, captq);

    /* Unlink all buffers, except the one on head. We're currently grabbing
     * into it, siusb_buf_complete() will unlink it.
     */
    list_for_each_entry_safe(buf, tmp_buf, &s->captq, captq) {
        if (buf != active_buf)
            list_del(&buf->captq);
    }

    spin_unlock_irqrestore(&s->captq_lock, flags);

    /* If the frame transfer was not in init phase, but currently running bulk
     * transfers, we'll wait for bulks to complete.
     */
    result = siusb_readout_wait(s, active_buf);
    if (result < 0)
        return result;

    return 0;
}



/*
 * Resource allocation
 */

static void siusb_free_buf(struct siusb_buf *buf)
{
    unsigned j;

    if (buf->pages) {
        for (j = 0; j < siusb_pages_per_buf; j++) {
            if (buf->pages[j])
                free_page(buf->pages[j]);
        }

        kfree(buf->pages);
        buf->pages = NULL;
    }
}

static int siusb_alloc_buf(struct siusb_buf *buf, unsigned nr)
{
    unsigned j;

    buf->nr = nr;
    buf->state = SIUSB_BUF_UNUSED;
    init_waitqueue_head(&buf->done);

    buf->pages = kmalloc(siusb_pages_per_buf * sizeof(*buf->pages), GFP_KERNEL);
    if (!buf->pages) {
        printk(KERN_ERR "siusb: buf#%u: kmalloc(pages): FAILED\n", buf->nr);
        goto err;
    }

    memset(buf->pages, 0, siusb_pages_per_buf * sizeof(*buf->pages));

    for (j = 0; j < siusb_pages_per_buf; j++) {
        buf->pages[j] = __get_free_page(GFP_KERNEL);
        if (!buf->pages[j]) {
            printk(KERN_ERR "siusb: buf#%u: __get_free_page(pages[%u]): FAILED\n", buf->nr, j);
            goto err;
        }
    }

    return 0;

err:
    siusb_free_buf(buf);
    return -ENOMEM;
}

static void siusb_free_read_buf(struct siusb *s)
{
    siusb_free_buf(&s->read_buf);
}

static int siusb_alloc_read_buf(struct siusb *s)
{
    return siusb_alloc_buf(&s->read_buf, num_bufs);
}

static void siusb_free_bufs(struct siusb *s)
{
    unsigned i;

    if (s->bufs) {
        for (i = 0; i < num_bufs; i++)
            siusb_free_buf(&s->bufs[i]);

        kfree(s->bufs);
        s->bufs = NULL;

        TRACE1("siusb: buffers released\n");
    }
}

static int siusb_alloc_bufs(struct siusb *s)
{
    unsigned i;
    int result;

    s->bufs = kmalloc(num_bufs * sizeof(*s->bufs), GFP_KERNEL);
    if (!s->bufs) {
        printk(KERN_ERR "siusb: kmalloc(bufs): FAILED\n");
        result = -ENOMEM;
        goto err;
    }

    memset(s->bufs, 0, num_bufs * sizeof(*s->bufs));

    for (i = 0; i < num_bufs; i++) {
        result = siusb_alloc_buf(&s->bufs[i], i);
        if (result < 0)
            goto err;
    }

    TRACE1("siusb: buffers allocated\n");
    return 0;

err:
    siusb_free_bufs(s);
    return result;
}


/*
 * V4L interface
 */

static int siusb_set_size(struct siusb *s, unsigned x, unsigned y, unsigned w, unsigned h)
{
    TRACE1("siusb: %s: x=%u,y=%u,w=%u,h=%u\n", __func__, x, y, w, h);

    x = (x/2)*2;
    w = (w/2)*2;

    if (w < SIUSB_MIN_WIDTH)
        return -EINVAL;
    if (w > SIUSB_MAX_WIDTH)
        w = SIUSB_MAX_WIDTH;

    if (h < SIUSB_MIN_HEIGHT)
        return -EINVAL;
    if (h > SIUSB_MAX_HEIGHT)
        h = SIUSB_MAX_HEIGHT;

    if (x + w > SIUSB_MAX_WIDTH || y + h > SIUSB_MAX_HEIGHT)
        return -EINVAL;

    s->x = x;
    s->y = y;
    s->w = w;
    s->h = h;
    return 0;
}

static int siusb_open(struct inode *inode, struct file *file)
{
    struct video_device *vdev;
    struct siusb *s;
    int result;

    result = down_interruptible(&disconnect_sem);   /* Prevent disconnects */
    if (result < 0)
        return result;

    vdev = video_devdata(file);
    if (!vdev) {
        up(&disconnect_sem);
        return -ENODEV;
    }

    s = video_get_drvdata(vdev);

    result = down_interruptible(&s->sem);
    if (result < 0) {
        up(&disconnect_sem);
        return result;
    }

    /* Allow single open.
     */
    if (s->open_count > 0) {
        result = -EBUSY;
        goto out;
    }

    /* First open().
     */
    result = siusb_alloc_read_buf(s);   /* Allocate read() syscall buffer */
    if (result < 0)
        goto out;

    s->open_count++;
    file->private_data = s;

    /* Start read() syscall streaming.
     */
    s->switched_to_mmap = 0;
    siusb_capture_request(s, &s->read_buf);

out:
    up(&s->sem);
    up(&disconnect_sem);
    return result;
}

static int siusb_release(struct inode *inode, struct file *file)
{
    struct siusb *s = file->private_data;
    int result;

    result = down_interruptible(&s->sem);
    if (result < 0)
        return result;

    result = siusb_cancel(s);
    if (result < 0)
        goto out;

    s->open_count--;

    if (s->open_count <= 0) {
        /* Last release().
         */
        siusb_free_bufs(s);
        siusb_free_read_buf(s);
    }

    if (!s->present) {
        /* The camera was unplugged before the file was released.
         */
        up(&s->sem);
        kfree(s);
        return 0;
    }

out:
    up(&s->sem);
    return result;
}

static ssize_t siusb_read(struct file *file, char *ptr, size_t count, loff_t *ppos)
{
    struct siusb *s = file->private_data;
    int rem;
    unsigned cpage;
    int result;

    result = down_interruptible(&s->sem);
    if (result < 0)
        return result;

    /* Verify that the camera wasn't unplugged.
     */
    if (!s->present) {
        result = -ENODEV;
        goto out;
    }

    /* Once the user has switched away from using the read() syscall, he can't
     * return back to using it without reopen.
     */
    if (s->switched_to_mmap) {
        result = -EINVAL;
        goto out;
    }

    result = siusb_readout_wait(s, &s->read_buf);
    if (result < 0) {
        /* Interrupted by signal, i.e. buffer not synced yet, so we shouldn't
         * start new capture.
         */
        goto out;
    }

    if (s->read_buf.result < 0) {
        result = s->read_buf.result;
        goto out_capture_request;
    }

    if (count > s->read_buf.w * s->read_buf.h)
        count = s->read_buf.w * s->read_buf.h;

    for (rem = count, cpage = 0; rem > 0; rem -= PAGE_SIZE, cpage++) {
        if (copy_to_user(ptr, (void *)s->read_buf.pages[cpage], min(rem, (int)PAGE_SIZE))) {
            result = -EFAULT;
            goto out_capture_request;
        }

        ptr += PAGE_SIZE;
    }

    result = count;

out_capture_request:
    siusb_capture_request(s, &s->read_buf);
out:
    up(&s->sem);
    return result;
}

static unsigned siusb_poll_buf(struct file *file, struct poll_table_struct *wait,
                               struct siusb_buf *buf)
{
    if (buf->state != SIUSB_BUF_UNUSED) {
        poll_wait(file, &buf->done, wait);
        if (buf->state == SIUSB_BUF_DONE) {
            if (buf->result < 0) {
                /* TODO: Support for POLLERR.
                 */
            }
            return POLLIN | POLLRDNORM;
        }
    }
    return 0;
}

static unsigned siusb_poll(struct file *file, struct poll_table_struct *wait)
{
    struct siusb *s = file->private_data;
    unsigned i;
    unsigned mask;

    /* TODO:
     *
     * Think about races here.
     *
     * Should we first add ALL wait queues and after that check ready
     * conditions?
     */

    if (s->switched_to_mmap) {
        for (i = 0; i < num_bufs; i++) {
            mask = siusb_poll_buf(file, wait, &s->bufs[i]);
            if (mask)
                return mask;
        }
    } else {
        mask = siusb_poll_buf(file, wait, &s->read_buf);
        if (mask)
            return mask;
    }

    return 0;
}

static struct page *siusb_vma_nopage(struct vm_area_struct *vma, unsigned long address,
#ifdef LINUX_24
int unused
#else
int *type
#endif
)
{
    struct siusb *s = vma->vm_private_data;
    unsigned page_offset;   /* Faulting page number in this VMA */
    unsigned buf_nr;        /* Buffer number, the page belongs to */
    unsigned page_nr;       /* Page number in that buffer */
    void *pageptr;
    struct page *page;

    page_offset = vma->vm_pgoff + (address - vma->vm_start) / PAGE_SIZE;

    if (page_offset >= num_bufs * siusb_pages_per_buf)  /* Out of range */
        return NOPAGE_SIGBUS;

    buf_nr = page_offset / siusb_pages_per_buf;
    page_nr = page_offset % siusb_pages_per_buf;

    pageptr = (void *)s->bufs[buf_nr].pages[page_nr];
    page = virt_to_page(pageptr);

    get_page(page);
    return page;
}

static struct vm_operations_struct siusb_vm_ops = {
    .nopage =   siusb_vma_nopage,
};

static int siusb_mmap(struct file *file, struct vm_area_struct *vma)
{
    struct siusb *s = file->private_data;
    int result;

    result = down_interruptible(&s->sem);
    if (result < 0)
        return result;

    /* Verify that the camera wasn't unplugged.
     */
    if (!s->present) {
        result = -ENODEV;
        goto out;
    }

    /* Don't allow mmap() before buffers have been requested (through
     * VIDIOCGMBUF ioctl).
     */
    if (!s->switched_to_mmap) {
        result = -EINVAL;
        goto out;
    }

    /* Don't do anything here, nopage() will fill the holes.
     */
    vma->vm_flags |= VM_RESERVED;
    vma->vm_ops = &siusb_vm_ops;
    vma->vm_private_data = s;

out:
    up(&s->sem);
    return result;
}

static int siusb_do_ioctl(struct inode *inode, struct file *file, unsigned cmd, void *arg)
{
    struct siusb *s = file->private_data;
    int result;

    result = down_interruptible(&s->sem);
    if (result < 0)
        return result;

    /* Verify that the camera wasn't unplugged.
     */
    if (!s->present) {
        result = -ENODEV;
        goto out;
    }

    switch (cmd) {
        case VIDIOCGCAP:
        {
            struct video_capability *cap = arg;

            strcpy(cap->name, s->version);
            cap->type = VID_TYPE_CAPTURE | VID_TYPE_SUBCAPTURE | VID_TYPE_MONOCHROME;
            cap->channels = 1;
            cap->audios = 0;
            cap->maxwidth = SIUSB_MAX_WIDTH;
            cap->maxheight = SIUSB_MAX_HEIGHT;
            cap->minwidth = SIUSB_MIN_WIDTH;
            cap->minheight = SIUSB_MIN_HEIGHT;
            break;
        }

        case VIDIOCGCHAN:
        {
            struct video_channel *chan = arg;

            if (chan->channel != 0) {
                result = -EINVAL;
                break;
            }
            strcpy(chan->name, "Camera");
            chan->tuners = 0;
            chan->flags = 0;
            chan->type = VIDEO_TYPE_CAMERA;
            chan->norm = VIDEO_MODE_AUTO;
            break;
        }

        case VIDIOCSCHAN:
        {
            struct video_channel *chan = arg;

            if (chan->channel != 0 || chan->norm != VIDEO_MODE_AUTO)
                result = -EINVAL;
            break;
        }

        case VIDIOCGPICT:
        {
            struct video_picture *pict = arg;

            pict->brightness = s->exposure << 4;
            pict->hue = 32768;
            pict->colour = 32768;
            pict->contrast = 32768;
            pict->whiteness = 32768;
            pict->depth = 8;
            pict->palette = VIDEO_PALETTE_GREY;
            break;
        }

        case VIDIOCSPICT:
        {
            struct video_picture *pict = arg;

            if (pict->depth != 8 || pict->palette != VIDEO_PALETTE_GREY) {
                result = -EINVAL;
                break;
            }
            s->exposure = pict->brightness >> 4;
            break;
        }

        case VIDIOCGWIN:
        {
            struct video_window *win = arg;

            win->x = 0;
            win->y = 0;
            win->width = s->w;
            win->height = s->h;
            win->chromakey = 0;
            win->flags = 0;
            win->clips = NULL;
            win->clipcount = 0;
            break;
        }

        case VIDIOCSWIN:
        {
            struct video_window *win = arg;

            if (win->clips != NULL || win->clipcount != 0) {
                result = -EINVAL;
                break;
            }

            result = siusb_set_size(s, (SIUSB_MAX_WIDTH - win->width) / 2,
                                    (SIUSB_MAX_HEIGHT - win->height) / 2,
                                    win->width, win->height);
            break;
        }

        case VIDIOCGCAPTURE:
        {
            struct video_capture *capt = arg;

            capt->x = s->x;
            capt->y = s->y;
            capt->width = s->w;
            capt->height = s->h;
            capt->decimation = 0;
            capt->flags = 0;
            break;
        }

        case VIDIOCSCAPTURE:
        {
            struct video_capture *capt = arg;

            if (capt->decimation != 0 || capt->flags != 0) {
                result = -EINVAL;
                break;
            }

            result = siusb_set_size(s, capt->x, capt->y, capt->width, capt->height);
            break;
        }

        case VIDIOCGMBUF:
        {
            struct video_mbuf *mbuf = arg;
            unsigned i;

            result = siusb_alloc_bufs(s);
            if (result < 0)
                break;

            /* We can't return to using read() method again, so read() buffer
             * may be released, but be sure we aren't capturing into it now.
             */
            result = siusb_cancel(s);
            if (result < 0) {
                siusb_free_bufs(s);
                break;
            }
            siusb_free_read_buf(s);

            s->switched_to_mmap = 1;

            mbuf->size = num_bufs * SIUSB_MAX_SIZE;
            mbuf->frames = num_bufs;
            for (i = 0; i < num_bufs; i++)
                mbuf->offsets[i] = i * SIUSB_MAX_SIZE;
            break;
        }

        case VIDIOCMCAPTURE:
        {
            struct video_mmap *mmap = arg;

            if (!s->switched_to_mmap) {
                result = -EINVAL;
                break;
            }

            if (mmap->frame >= num_bufs) {
                TRACE1("siusb: VIDIOCMCAPTURE: bad frame=%u\n", mmap->frame);
                result = -EINVAL;
                break;
            }

            if (mmap->format != VIDEO_PALETTE_GREY) {
                TRACE1("siusb: VIDIOCMCAPTURE: bad format=%u\n", mmap->format);
                result = -EINVAL;
                break;
            }

            if (s->bufs[mmap->frame].state != SIUSB_BUF_UNUSED) {
                result = -EBUSY;
                break;
            }

            /* Set new size only if different to prevent image centering if the
             * user has already set preferred origin and size.
             */
            if (s->w != mmap->width || s->h != mmap->height) {
                result = siusb_set_size(s, (SIUSB_MAX_WIDTH - mmap->width) / 2,
                                           (SIUSB_MAX_HEIGHT - mmap->height) / 2,
                                           mmap->width, mmap->height);
                if (result < 0)
                    break;
            }

            siusb_capture_request(s, &s->bufs[mmap->frame]);
            break;
        }

        case VIDIOCSYNC:
        {
            unsigned *frame = arg;

            if (!s->switched_to_mmap) {
                result = -EINVAL;
                break;
            }

            if (*frame >= num_bufs) {
                result = -EINVAL;
                break;
            }

            result = siusb_readout_wait(s, &s->bufs[*frame]);
            if (result < 0)
                break;

            /* TODO: what should return VIDIOCSYNC if there is frame capture error?
             *
             * We return -EINVAL on bad arg
             *           -EINTR if signaled (probably)
             * These two mean the frame is still NOT finished and the user should reinvoke
             * VIDIOCSYNC.
             *
             * The other case is the frame is finished, but with an error -- we return
             * the error code!
             *
             * So on error return how does the user find out if the VIDIOCSYNC call must
             * be reinvoked?
             */
            result = s->bufs[*frame].result;
            break;
        }

        case VIDIOCGUNIT:
        {
            struct video_unit *unit = arg;

            unit->video = s->vdev->minor;
            unit->vbi = VIDEO_NO_UNIT;
            unit->radio = VIDEO_NO_UNIT;
            unit->audio = VIDEO_NO_UNIT;
            unit->teletext = VIDEO_NO_UNIT;
            break;
        }

        /* No overlay.
         */
        case VIDIOCCAPTURE:
        case VIDIOCGFBUF:
        case VIDIOCSFBUF:
        case VIDIOCKEY:

        /* No tuner.
         */
        case VIDIOCGTUNER:
        case VIDIOCSTUNER:
        case VIDIOCGFREQ:
        case VIDIOCSFREQ:

        /* No audio.
         */
        case VIDIOCGAUDIO:
        case VIDIOCSAUDIO:
            result = -EINVAL;
            break;

        case SIUSB_IOC_GET_CLOCK:
        {
            unsigned *clk = arg;

            *clk = s->clock;
            break;
        }

        case SIUSB_IOC_SET_CLOCK:
        {
            unsigned *clk = arg;

            s->clock = *clk;
            break;
        }

        case SIUSB_IOC_GET_SHUTTER_MODE:
        {
            unsigned *mode = arg;

            *mode = s->shutter_mode;
            break;
        }

        case SIUSB_IOC_SET_SHUTTER_MODE:
        {
            unsigned *mode = arg;

            s->shutter_mode = *mode;
            break;
        }

        case SIUSB_IOC_GET_EXPOSURE:
        {
            unsigned *exp = arg;

            *exp = s->exposure;
            break;
        }

        case SIUSB_IOC_SET_EXPOSURE:
        {
            unsigned *exp = arg;

            /* TODO: understand how this works and what exp actually means for the
             * different TBs/shutter modes.
             */
            /* TODO:
             *
             *  Shutter mode = TRIG_TB3 -- the first image line is bad. This seems not to
             *  happen on TB4.
             *
             *  Check if passed val is within range.
             */
            /* TODO: It seems like first image line is *always* bad, I just haven't seen it
             * on TB4, because it's too dark.
             */

            if (*exp > SIUSB_EXPOSURE_MAX) {
                result = -EINVAL;
                break;
            }
            s->exposure = *exp;
            break;
        }

        case SIUSB_IOC_GET_GAIN:
        {
            unsigned *gain = arg;

            *gain = s->gain;
            break;
        }

        case SIUSB_IOC_SET_GAIN:
        {
            unsigned *gain = arg;

            if (*gain < SIUSB_GAIN_MIN || *gain > SIUSB_GAIN_UNITY) {
                result = -EINVAL;
                break;
            }
            s->gain = *gain;
            break;
        }

        case SIUSB_IOC_GET_BLACK_LEVEL:
        {
            unsigned *val = arg;

            *val = s->black_level;
            break;
        }

        case SIUSB_IOC_SET_BLACK_LEVEL:
        {
            unsigned *val = arg;

            if (*val > SIUSB_BLACK_LEVEL_MAX) {
                result = -EINVAL;
                break;
            }
            s->black_level = *val;
            break;
        }

        case SIUSB_IOC_GET_BLACK_BALANCE:
        {
            unsigned *val = arg;

            *val = s->black_balance;
            break;
        }

        case SIUSB_IOC_SET_BLACK_BALANCE:
        {
            unsigned *val = arg;

            if (*val > SIUSB_BLACK_BALANCE_MAX) {
                result = -EINVAL;
                break;
            }
            s->black_balance = *val;
            break;
        }

        case SIUSB_IOC_GET_TRIG_MODE:
        {
            unsigned *mode = arg;

            *mode = s->trig_mode;
            break;
        }

        case SIUSB_IOC_SET_TRIG_MODE:
        {
            unsigned *mode = arg;

            s->trig_mode = *mode;
            break;
        }

        default:
            result = -ENOIOCTLCMD;
            break;
    }

out:
    up(&s->sem);
    return result;
}

static int siusb_ioctl(struct inode *inode, struct file *file, unsigned cmd, unsigned long arg)
{
    return video_usercopy(inode, file, cmd, arg, siusb_do_ioctl);
}

static struct file_operations siusb_fops = {
    .owner =    THIS_MODULE,
    .open =     siusb_open,
    .release =  siusb_release,
    .read =     siusb_read,
    .poll =     siusb_poll,
    .mmap =     siusb_mmap,
    .ioctl =    siusb_ioctl,
    .llseek =   no_llseek,
};

static struct video_device siusb_vdev_template = {
    .owner =    THIS_MODULE,
    .type =     VID_TYPE_CAPTURE,
    .hardware = VID_HARDWARE_SI,
    .fops =     &siusb_fops,
    .release =  video_device_release,
};

static int siusb_v4l_register(struct siusb *s)
{
    int result;

    s->vdev = video_device_alloc();
    if (!s->vdev) {
        printk(KERN_ERR "siusb: video_device_alloc(): FAILED\n");
        return -ENOMEM;
    }

    *s->vdev = siusb_vdev_template;
    strcpy(s->vdev->name, s->version);
    video_set_drvdata(s->vdev, s);

    result = video_register_device(s->vdev, VFL_TYPE_GRABBER, -1);
    if (result < 0) {
        printk(KERN_ERR "siusb: video_register_device()=%d: FAILED\n", result);
        video_device_release(s->vdev);
        return result;
    }

    return 0;
}

static void siusb_v4l_unregister(struct siusb *s)
{
    video_unregister_device(s->vdev);
}


/*
 * Initialization
 */

static int siusb_reset_cam(struct siusb *s)
{
    int result;

    result = siusb_set_depth(s, SIUSB_DEPTH_8BIT);
    if (result < 0)
        return result;

    /* Initialize current requested and hardware settings with different values.
     * On the first frame capture, hardware settings will be set to match requested.
     */

    s->x = 0;
    s->hw_x = 2;

    s->y = 0;
    s->hw_y = 2;

    s->w = SIUSB_MAX_WIDTH;
    s->hw_w = SIUSB_MIN_WIDTH;

    s->h = SIUSB_MAX_HEIGHT;
    s->hw_h = SIUSB_MIN_HEIGHT;

    s->clock = SIUSB_CLK_20MHZ;
    s->hw_clock = SIUSB_CLK_25MHZ;

    s->shutter_mode = SIUSB_SH_TRIG_TB3;
    s->hw_shutter_mode = SIUSB_SH_TRIG_TB1;

    s->exposure = 0x800;
    s->hw_exposure = 0x100;

    s->gain = SIUSB_GAIN_1_88;
    s->hw_gain = SIUSB_GAIN_1_41;

    s->black_level = 85;
    s->hw_black_level = 10;

    s->black_balance = 85;
    s->hw_black_balance = 10;

    return 0;
}

static struct siusb *siusb_cam;

static int siusb_probe(void)
{
    struct siusb *s;
    int result;

    /* Allocate memory for our device.
     */
    s = kmalloc(sizeof(*s), GFP_KERNEL);
    if (!s) {
        printk(KERN_ERR "siusb: %s: kmalloc() failed\n", __func__);
        result = -ENOMEM;
        goto err;
    }

    /* And initialize it.
     */
    memset(s, 0, sizeof(*s));
    s->present = 1;         /* Allow file operations */
    init_MUTEX(&s->sem);

    /* Reset camera.
     */
    result = siusb_reset_cam(s);
    if (result < 0)
        goto err;

    /* Read camera firmware version.
     */
    result = siusb_read_version(s, s->version);
    if (result < 0)
        goto err;

    /* Init capture queue.
     */
    INIT_LIST_HEAD(&s->captq);
    spin_lock_init(&s->captq_lock);

    result = siusb_v4l_register(s);
    if (result < 0)
        goto err;

    siusb_cam = s;
    printk(KERN_INFO "siusb: %s attached\n", s->version);
    return 0;

err:
    kfree(s);
    return result;
}

static void siusb_disconnect(void)
{
    struct siusb *s;

    down(&disconnect_sem);

    s = siusb_cam;

    down(&s->sem);
    s->present = 0;             /* Prevent file operations */
    siusb_v4l_unregister(s);
    up(&s->sem);

    /* If the device is opened, siusb_release() will clean this up.
     */
    if (s->open_count <= 0)
        kfree(s);

    up(&disconnect_sem);
    printk(KERN_INFO "siusb: disconnected\n");
}

static int __init siusb_init(void)
{
    int result;

    BUG_ON(SIUSB_MAX_SIZE % PAGE_SIZE != 0);    /* We rely on this */
    siusb_pages_per_buf = SIUSB_MAX_SIZE / PAGE_SIZE;

    result = siusb_probe();
    if (result < 0)
        return result;

    printk(KERN_INFO DRIVER_DESC " " DRIVER_VERSION "\n");
    return 0;
}

static void __exit siusb_exit(void)
{
    siusb_disconnect();
}

module_init(siusb_init);
module_exit(siusb_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("Proprietary");
